/*
 * A collection of functions that extend the GSL.
 *
 * Robert Johansson <robert@riken.jp>
 *
 * $Id: gsl_ext.c,v 1.1.1.1 2008-09-03 09:36:41 rob Exp $
 */

#include "math.h"

#include <gsl/gsl_complex.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_cblas.h>
#include <gsl/gsl_blas.h>

#include <blaswrap.h>
#include <f2c.h>
#include <clapack.h>


/*
 * Convert a real valued GSL matrix to a complex valued GSL matrix.
 *
 */
int
gsl_ext_matrix_convert(gsl_matrix_complex *dst, gsl_matrix *src)
{
	unsigned int i, j;
	gsl_complex z;
	
	if (src->size1 != dst->size1 || src->size2 != dst->size2)
		return -1;

	for (i = 0; i < src->size1; i++)
	{
		for (j = 0; j < src->size1; j++)
		{
			GSL_SET_COMPLEX(&z, gsl_matrix_get(src, i, j), 0.0);
			gsl_matrix_complex_set(dst, i, j, z);
		}
	}
	
	return 0;
}

/* 
 * Allocate a complex matrix, convert the real matrix it complex form
 * and free the real matrix.
 *
 */
gsl_matrix_complex *
gsl_ext_matrix_convert_and_free(gsl_matrix *m)
{
	gsl_matrix_complex *zm;

	zm = gsl_matrix_complex_alloc(m->size1, m->size2);

	gsl_ext_matrix_convert(zm, m);

	gsl_matrix_free(m);
	
	return zm;
}

/*
 * Calculate the matrix exponent of A and store in eA.
 * Algorithm: Truncated Talyor series.
 *
 * WARNING: Large errors possible and it's slow.
 */
int 
gsl_ext_expm_complex(gsl_matrix_complex *A, gsl_matrix_complex *eA)
{
	int i;
	gsl_complex alpha, beta, z;
	gsl_matrix_complex *I, *T;
	
	I = gsl_matrix_complex_alloc(A->size1, A->size2);
	T = gsl_matrix_complex_alloc(A->size1, A->size2);
	
	GSL_SET_COMPLEX(&alpha, 1.0, 0.0);
	GSL_SET_COMPLEX(&beta,  0.0, 0.0);
	
	gsl_matrix_complex_set_identity(I);
	gsl_matrix_complex_set_identity(eA);
	
	for (i = 50; i > 0; i--)
	{
		GSL_SET_COMPLEX(&z, 1.0 / i, 0.0);
		gsl_matrix_complex_scale(eA, z);
		
		gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, eA, A, beta, T);	
		gsl_matrix_complex_add(T, I);
		gsl_matrix_complex_memcpy(eA, T);
	}
	
	return 0;
}

/*
 * Calculate eigenvectors and eigenvalues for a non-symmetric complex matrix
 * using the CLAPACK zgeev_ function.
 *
 */
int
gsl_ext_eigen_zgeev(gsl_matrix_complex *A_gsl, gsl_matrix_complex *evec, gsl_vector_complex *eval)
{
        integer n,i,j,*pivot,info,lwork, ldvl, ldvr, lda;
        doublecomplex *A,*vr,*vl,*w,*work;
        doublereal *rwork;
        char jobvl,jobvr;

	n = A_gsl->size1;
	
        pivot = (integer *)malloc((size_t)n * sizeof(int));
        A  = (doublecomplex *)malloc((size_t)n * n * sizeof(doublecomplex));
        w  = (doublecomplex *)malloc((size_t)n * sizeof(doublecomplex));
        vr = (doublecomplex *)malloc((size_t)n * n * sizeof(doublecomplex));
        vl = (doublecomplex *)malloc((size_t)n * n * sizeof(doublecomplex));
        lwork = 10 * n;
        work  = (doublecomplex *)malloc((size_t)lwork * sizeof(doublecomplex));
        rwork  = (doublereal *)malloc((size_t)lwork * sizeof(doublereal));

        for (i = 0; i < n; i++)
        {
                for (j = 0; j < n; j++)
                {
			gsl_complex z;
                        double re,im;
		
			z = gsl_matrix_complex_get(A_gsl, i, j);
			re = GSL_REAL(z);
			im = GSL_IMAG(z);

                        A[j*n+i] = (doublecomplex){re,im};
                }
        }

        jobvl='N';
        jobvr='V';

        lda  = n;
        ldvr = n;
        ldvl = n;
        zgeev_(&jobvl, &jobvr, &n, A, &lda, w, vl, &ldvl, vr, &ldvr, work, &lwork, rwork, &info);

        for (i = 0; i < n; i++)
        {
		gsl_complex z;
		GSL_SET_COMPLEX(&z, w[i].r, w[i].i);
		gsl_vector_complex_set(eval, i, z);
        }

	for (j = 0; j < n; j++)
	{
        	for (i = 0; i < n; i++)
        	{
			gsl_complex z;
			
			GSL_SET_COMPLEX(&z, vr[j*n+i].r, vr[j*n+i].i);
			
			gsl_matrix_complex_set(evec, i, j, z);
        	}
	}

	printf("zgeev_: info = %d\n", (int)info);
	
	free(pivot);
	free(A);
	free(w);
	free(vr);
	free(vl);
	free(work);
	free(rwork);
	
	return 0;
}

