/*
 * Functions for generating hamiltonians in matrix form. 
 * 
 * Robert Johansson <robert@riken.jp>
 *
 * $Id: hamiltonian.c,v 1.4 2008-10-23 07:15:47 rob Exp $
 */

#include <stdio.h>
#include <math.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_complex.h>
#include <gsl/gsl_complex_math.h>

#include "quantum_system.h"
#include "operators.h"
#include "hamiltonian.h"
#include "param.h"
#include "gsl_ext.h"


/**
 * \breif	Generate the hamiltonian at time t.
 *
 * For cases where the time-dependence of the Hamiltonian takes the form
 * H = H0 + H1 * f(t), this function is used to implement the specific
 * time-dependence in f(t).
 *
 * @param	H_t	Matrix where the resulting time-dependent Hamiltonian
 * 			at time t should be stored (output from this function).
 * @param	H0	The constant part of the Hamiltonian.
 * @param	H1	The part of the Hamiltonian which is modulated by f(t).
 * @param	t	The instant of time for which we should calculate H(t).
 * @param	p	Pointer to the data structure that contains all the 
 * 			simulation parameters.
 */
int
hamiltonian_t(gsl_matrix_complex *H_t, gsl_matrix_complex *H0, gsl_matrix_complex *H1, double t, solver_param *p)
{
	if (p->f_A == 0.0)
	{
		gsl_matrix_complex_memcpy(H_t, H0);

		// p->H_t = p->H_0;
	}
	else
	{
		gsl_complex z;

		// f(t) = A * sin(w * t)

		gsl_matrix_complex_memcpy(H_t, H0);
		gsl_matrix_complex_memcpy(H_t, H1);
		GSL_SET_COMPLEX(&z, p->f_A * sin(p->f_w * t), 0.0);
		gsl_matrix_complex_scale(H_t, z);
		gsl_matrix_complex_add(H_t, H0);
	}
	
	return 0;
}

/**
 * \breif	Generates the hamiltonian for the driving field.
 *
 * Often time-dependet hamiltonians take the form H = H_0 + H_drive * f(t).
 * This function should generate the H_drive part of the total Hamiltonian.
 * @see hamiltonian_t.
 * 
 * @param	qs	Pointer to the data structure that defines
 * 			the arrangement of the quantum system.
 * @param	param	Pointer to a solver parameter structure
 * 			that includes all the parameter values
 * 			for the simulation.
 */
gsl_matrix_complex *
hamiltonian_drive(quantum_system *qs, solver_param *param)
{	
	int ns;
	gsl_matrix_complex *H_drive;

	ns = quantum_system_nstates(qs);

	H_drive = operator_sigma_x(qs, 0);
	
	return H_drive;
}

/*
 * Generate the hamiltonian for a N-TLS system
 *
 */
gsl_matrix_complex *
hamiltonian_nTLS(quantum_system *qs, solver_param *param)
{
        int ns, i, j;
        gsl_matrix_complex *H, *s1, *s2, *s3;

        ns = quantum_system_nstates(qs);

	printf("hamiltonian_nTLS: DEBUG: Generating nTLS hamiltonian with eps = %f, %f, ...\n", param->epsilon[0], param->epsilon[1]);

        if ((H = gsl_matrix_complex_alloc(ns, ns)) == NULL)
        {
                fprintf(stderr, "hamiltonian_nTLS: Failed to allocate H. Out of memory?\n");
                return NULL;
        }

        gsl_matrix_complex_set_zero(H);

        s3 = gsl_matrix_complex_alloc(ns,ns);

        /*
         * the TLS energies
         */
        for (i = 0; i < qs->nsubsys; i++)
        {
		gsl_complex z;

                s1 = operator_sigma_z(qs, i);
		GSL_SET_COMPLEX(&z, -0.5 * param->epsilon[i], 0.0);
                gsl_matrix_complex_scale(s1, z);
                gsl_matrix_complex_add(H, s1);
                gsl_matrix_complex_free(s1);

                s1 = operator_sigma_x(qs, i);
		GSL_SET_COMPLEX(&z, -0.5 * param->delta[i], 0.0);
                gsl_matrix_complex_scale(s1, z);
                gsl_matrix_complex_add(H, s1);
                gsl_matrix_complex_free(s1);
        }

        /*
         * TLS interactions
         */
        for (i = 0; i < qs->nsubsys; i++)
        {
                for (j = 0; j < qs->nsubsys; j++)
                {
                        if (param->lambda[i][j] != 0.0)
                        {
				gsl_complex z;

                                s1 = operator_sigma_x(qs, i);
                                s2 = operator_sigma_x(qs, j);

				GSL_SET_COMPLEX(&z, -0.5 * param->lambda[i][j], 0.0);
                                gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, z, s1, s2, GSL_COMPLEX_ZERO, s3);
                                gsl_matrix_complex_add(H, s3);

                                gsl_matrix_complex_free(s1);
                                gsl_matrix_complex_free(s2);
                        }
                }
        }

        gsl_matrix_complex_free(s3);

        return H;
}


/*
 * Generate the hamiltonian for a N-TLS system 
 *
 */
gsl_matrix_complex *
hamiltonian_spinchain(quantum_system *qs, solver_param *param)
{
	int ns, i, j;
	gsl_matrix_complex *H, *s1, *s2, *s3, *s1b, *s2b;

	ns = quantum_system_nstates(qs);
	
	if ((H = gsl_matrix_complex_alloc(ns, ns)) == NULL)
	{
		fprintf(stderr, "hamiltonian_nTLS: Failed to allocate H. Out of memory?\n");
		return NULL;
	}

	gsl_matrix_complex_set_zero(H);
	
	s3 = gsl_matrix_complex_alloc(ns,ns);

	/*
	 * the TLS energies
	 */
	for (i = 0; i < qs->nsubsys; i++)
	{
		gsl_complex z;
		s1 = operator_sigma_z(qs, i);
		GSL_SET_COMPLEX(&z, -0.5 * param->epsilon[i], 0.0);
		gsl_matrix_complex_scale(s1, z);
		gsl_matrix_complex_add(H, s1);
		gsl_matrix_complex_free(s1);

		
		s1 = operator_sigma_x(qs, i);
		GSL_SET_COMPLEX(&z, -0.5 * param->delta[i], 0.0);
		gsl_matrix_complex_scale(s1, z);
		gsl_matrix_complex_add(H, s1);
		gsl_matrix_complex_free(s1);
	}

	/*
	 * TLS interactions
	 */
	for (i = 0; i < qs->nsubsys; i++)
	{
		for (j = 0; j < qs->nsubsys; j++)
		{
			if (param->lambda[i][j] != 0.0)
			{
				gsl_complex z;
				double theta;
				
				s1 = operator_sigma_x(qs, i);
				s1b = operator_sigma_z(qs, i);
				s2 = operator_sigma_x(qs, j);
				s2b = operator_sigma_z(qs, j);
				s3 = gsl_matrix_complex_alloc(ns,ns);
				
				theta = tan(param->delta[i] / param->epsilon[i]);
				GSL_SET_COMPLEX(&z, sin(theta), 0.0);
				gsl_matrix_complex_scale(s1, z);
				GSL_SET_COMPLEX(&z, cos(theta), 0.0);
				gsl_matrix_complex_scale(s1b, z);
				gsl_matrix_complex_add(s1, s1b);
				theta = tan(param->delta[j] / param->epsilon[j]);
				GSL_SET_COMPLEX(&z, sin(theta), 0.0);
				gsl_matrix_complex_scale(s2, z);
				GSL_SET_COMPLEX(&z, cos(theta), 0.0);
				gsl_matrix_complex_scale(s2b, z);
				gsl_matrix_complex_add(s2, s2b);
				
				GSL_SET_COMPLEX(&z, -0.5 * param->lambda[i][j], 0.0);
				gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, z, s1, s2, GSL_COMPLEX_ZERO, s3);
				
				gsl_matrix_complex_add(H, s3);
				
				gsl_matrix_complex_free(s1);
				gsl_matrix_complex_free(s1b);
				gsl_matrix_complex_free(s2);
				gsl_matrix_complex_free(s2b);
			}
		}
	}

	gsl_matrix_complex_free(s3);
	
	return H;
}



/**
 * \breif	Generate the hamiltonian for two coupled harmonic oscillators
 *
 * This function allocate and initiates a matrix representation of a hamiltonian
 * of two coupled harmonic oscillators.
 *
 * @param	qs	Data structure that defines the quantum system.
 * @param	param	Data structure that contains all the parameters.
 *
 */
gsl_matrix_complex *
hamiltonian_2ho(quantum_system *qs, solver_param *param)
{
	int ns;
	gsl_matrix_complex *H, *a, *ad, *b, *bd, *ws, *Hint;
	gsl_complex alpha, beta, z;
	
	ns = quantum_system_nstates(qs);
	
	if ((H = gsl_matrix_complex_alloc(ns, ns)) == NULL || (Hint = gsl_matrix_complex_alloc(ns, ns)) == NULL)
	{
		fprintf(stderr, "hamiltonian_nTLS: Failed to allocate H and Hint. Out of memory?\n");
		return NULL;
	}

	gsl_matrix_complex_set_zero(H);
	//gsl_matrix_complex_set_zero(Hint);
	
	/* h_ w1 (ad a + 1/2) */
	a  = operator_ho_lowering(qs, 0, 0);
	ad = operator_ho_raising(qs,0, 0);
	GSL_SET_COMPLEX(&alpha, param->ho_w[0], 0.0);
	GSL_SET_COMPLEX(&beta, 0.0, 0.0);
	gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, ad, a, beta, H);
	ws = operator_unit(qs, -1);
	GSL_SET_COMPLEX(&z, 0.5 * param->ho_w[0], 0.0);
	gsl_matrix_complex_scale(ws, z);
	gsl_matrix_complex_add(H, ws);
	gsl_matrix_complex_free(ws);
	
	/* h_ w (bd b + 1/2) */
	b  = operator_ho_lowering(qs, 1, 0);
	bd = operator_ho_raising(qs, 1, 0);
	GSL_SET_COMPLEX(&alpha, param->ho_w[1], 0.0);
	GSL_SET_COMPLEX(&beta, 1.0, 0.0);
	gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, bd, b, beta, H);
	ws = operator_unit(qs, -1);
	GSL_SET_COMPLEX(&z, 0.5 * param->ho_w[1], 0.0);
	gsl_matrix_complex_scale(ws, z);
	gsl_matrix_complex_add(H, ws);
	gsl_matrix_complex_free(ws);
	
	// g (a + ad) (b + bd)
	gsl_matrix_complex_add(a, ad);
	gsl_matrix_complex_add(b, bd);
	GSL_SET_COMPLEX(&alpha, param->ho_g, 0.0);
	GSL_SET_COMPLEX(&beta, 1.0, 0.0);
	gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, a, b, beta, H);

	// g (a + ad) (b + bd)^2
	//gsl_matrix_complex_add(a, ad);
	//gsl_matrix_complex_add(b, bd);
	// g a * b^2
	//GSL_SET_COMPLEX(&alpha, 1.0, 0.0);
	//GSL_SET_COMPLEX(&beta,  0.0, 0.0);
	//gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, b, b, beta, Hint);
	//GSL_SET_COMPLEX(&alpha, param->ho_g, 0.0);
	//GSL_SET_COMPLEX(&beta,  1.0, 0.0);
	//gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, a, Hint, beta, H);
	//gsl_matrix_complex_scale(Hint, );
	//gsl_matrix_complex_add(H, Hint);
	//gsl_matrix_complex_free(Hint);


	// g (a + ad)^2 (b + bd)
	//gsl_matrix_complex_add(a, ad);
	//gsl_matrix_complex_add(b, bd);
	//gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, GSL_COMPLEX_ONE, a, a, GSL_COMPLEX_ZERO, Hint);
	//GSL_SET_COMPLEX(&alpha, param->ho_g, 0.0);
	//gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, Hint, b, GSL_COMPLEX_ONE, H);
	//gsl_matrix_complex_free(Hint);
	
	
	gsl_matrix_complex_free(a);
	gsl_matrix_complex_free(ad);
	gsl_matrix_complex_free(b);
	gsl_matrix_complex_free(bd);

	return H;
}


/**
 * Generate the hamiltonian for the 2-HO driving field
 *
 */
gsl_matrix_complex *
hamiltonian_2ho_drive(quantum_system *qs, solver_param *param)
{
	int ns;
	gsl_matrix_complex *H, *a, *ad; //, *b, *bd;
	
	ns = quantum_system_nstates(qs);
	
	if ((H = gsl_matrix_complex_alloc(ns, ns)) == NULL)
	{
		fprintf(stderr, "hamiltonian_nTLS: Failed to allocate H. Out of memory?\n");
		return NULL;
	}

	//gsl_matrix_complex_set_zero(H);
	
	a  = operator_ho_lowering(qs, 0, 0);
	ad = operator_ho_raising(qs, 0, 0);
	//b  = operator_ho_lowering(qs, 1);
	//bd = operator_ho_raising(qs, 1);

	// a bd + ad b
	//GSL_SET_COMPLEX(&alpha, 1.0, 0.0);
	//GSL_SET_COMPLEX(&beta,  0.0, 0.0);
	//gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, a, bd, beta, H);
	//GSL_SET_COMPLEX(&beta,  1.0, 0.0);
	//gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, ad, b, beta, H);
	
	// a + ad
	gsl_matrix_complex_set_zero(H);
	gsl_matrix_complex_add(H, a);
	gsl_matrix_complex_add(H, ad);
	

	gsl_matrix_complex_free(a);
	gsl_matrix_complex_free(ad);
	//gsl_matrix_complex_free(b);
	//gsl_matrix_complex_free(bd);
	
	return H;
}


/* ----------------------------------
 * Qubit + Cavity + Transmission line
 *
 */

/*
 * Generate the hamiltonian for the qubit + 2-HO driving field
 *
 */
gsl_matrix_complex *
hamiltonian_q2ho_drive(quantum_system *qs, solver_param *param)
{
	int ns;
	gsl_matrix_complex *H, *b, *bd;
	
	ns = quantum_system_nstates(qs);
	
	if ((H = gsl_matrix_complex_alloc(ns, ns)) == NULL)
	{
		fprintf(stderr, "hamiltonian_nTLS: Failed to allocate H. Out of memory?\n");
		return NULL;
	}
	
	b  = operator_ho_lowering(qs, 2, 0);
	bd = operator_ho_raising(qs, 2, 0);

	// a + ad
	gsl_matrix_complex_set_zero(H);
	gsl_matrix_complex_add(H, b);
	gsl_matrix_complex_add(H, bd);
	

	gsl_matrix_complex_free(b);
	gsl_matrix_complex_free(bd);
	
	return H;
}


/*
 * Generate the hamiltonian for the Qubit + Cavity + Transmission line system/
 *
 */
gsl_matrix_complex *
hamiltonian_q2ho(quantum_system *qs, solver_param *param)
{
	int ns;
	gsl_matrix_complex *H, *a, *ad, *b, *bd, *ws, *s1;
	gsl_complex alpha, z;
	
	ns = quantum_system_nstates(qs);
	
	if ((H = gsl_matrix_complex_alloc(ns, ns)) == NULL)
	{
		fprintf(stderr, "hamiltonian_nTLS: Failed to allocate H. Out of memory?\n");
		return NULL;
	}

	gsl_matrix_complex_set_zero(H);

	/* qubit */
        s1 = operator_sigma_z(qs, 0);
	GSL_SET_COMPLEX(&alpha, -0.5 * param->epsilon[0], 0.0);
        gsl_matrix_complex_scale(s1, alpha);
        gsl_matrix_complex_add(H, s1);
        gsl_matrix_complex_free(s1);

        s1 = operator_sigma_x(qs, 0);
        GSL_SET_COMPLEX(&alpha, -0.5 * param->delta[0], 0.0);
        gsl_matrix_complex_scale(s1, alpha);
        gsl_matrix_complex_add(H, s1);
        gsl_matrix_complex_free(s1);

	
	/* h_ w1 (ad a + 1/2) */
	a  = operator_ho_lowering(qs, 1, 0);
	ad = operator_ho_raising(qs,1, 0);
	GSL_SET_COMPLEX(&alpha, param->ho_w[0], 0.0);
	gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, ad, a, GSL_COMPLEX_ONE, H);
	ws = operator_unit(qs, -1);
	GSL_SET_COMPLEX(&z, 0.5 * param->ho_w[0], 0.0);
	gsl_matrix_complex_scale(ws, z);
	gsl_matrix_complex_add(H, ws);
	gsl_matrix_complex_free(ws);

	/* qubit - cavity interaction */
	s1 = operator_sigma_z(qs, 0);
	ws = gsl_matrix_complex_alloc(ns, ns);
	gsl_matrix_complex_memcpy(ws, a);
	gsl_matrix_complex_add(ws, ad);
	GSL_SET_COMPLEX(&alpha, param->g_qc, 0.0);
	gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, s1, ws, GSL_COMPLEX_ONE, H);
	gsl_matrix_complex_free(ws);
	gsl_matrix_complex_free(s1);
	

	/* h_ w (bd b + 1/2) */
	b  = operator_ho_lowering(qs, 2, 0);
	bd = operator_ho_raising(qs, 2, 0);
	GSL_SET_COMPLEX(&alpha, param->ho_w[1], 0.0);
	gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, bd, b, GSL_COMPLEX_ONE, H);
	ws = operator_unit(qs, -1);
	GSL_SET_COMPLEX(&z, 0.5 * param->ho_w[1], 0.0);
	gsl_matrix_complex_scale(ws, z);
	gsl_matrix_complex_add(H, ws);
	gsl_matrix_complex_free(ws);
		
	/* cavity - transmission line interaction : g_cr (a + ad) (b + bd) * /
	gsl_matrix_complex_add(a, ad);
	gsl_matrix_complex_add(b, bd);
	GSL_SET_COMPLEX(&alpha, param->g_cr, 0.0);
	gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, a, b, GSL_COMPLEX_ONE, H);
	*/
	
	gsl_matrix_complex_free(a);
	gsl_matrix_complex_free(ad);
//	gsl_matrix_complex_free(b);
//	gsl_matrix_complex_free(bd);

	return H;
}


/* -------------
 * Qubit + Cavity 
 *
 */

/*
 * Generate the hamiltonian for the qubit + HO + driving field
 *
 */
gsl_matrix_complex *
hamiltonian_qho_drive(quantum_system *qs, solver_param *param)
{
	int ns;
	gsl_matrix_complex *H, *b, *bd;
	
	ns = quantum_system_nstates(qs);
	
	if ((H = gsl_matrix_complex_alloc(ns, ns)) == NULL)
	{
		fprintf(stderr, "hamiltonian_nTLS: Failed to allocate H. Out of memory?\n");
		return NULL;
	}
	
	b  = operator_ho_lowering(qs, 1, param->n_offset);
	bd = operator_ho_raising(qs, 1, param->n_offset);

	// a + ad
	gsl_matrix_complex_set_zero(H);
	gsl_matrix_complex_add(H, b);
	gsl_matrix_complex_add(H, bd);
	

	gsl_matrix_complex_free(b);
	gsl_matrix_complex_free(bd);
	
	return H;
}

/*
 * Generate the hamiltonian for the Qubit + Cavity
 *
 */
gsl_matrix_complex *
hamiltonian_qho(quantum_system *qs, solver_param *param)
{
	int ns;
	gsl_matrix_complex *H, *a, *ad, *ws, *s1, *n;
	gsl_complex z, alpha;
	
	ns = quantum_system_nstates(qs);
	
	if ((H = gsl_matrix_complex_alloc(ns, ns)) == NULL)
	{
		fprintf(stderr, "hamiltonian_nTLS: Failed to allocate H. Out of memory?\n");
		return NULL;
	}

	gsl_matrix_complex_set_zero(H);

	/* qubit */
        s1 = operator_sigma_z(qs, 0);
	GSL_SET_COMPLEX(&alpha, -0.5 * param->epsilon[0], 0.0);
        gsl_matrix_complex_scale(s1, alpha);
        gsl_matrix_complex_add(H, s1);
        gsl_matrix_complex_free(s1);

        s1 = operator_sigma_x(qs, 0);
        GSL_SET_COMPLEX(&alpha, -0.5 * param->delta[0], 0.0);
        gsl_matrix_complex_scale(s1, alpha);
        gsl_matrix_complex_add(H, s1);
        gsl_matrix_complex_free(s1);

	
	/* h_ w1 (ad a + 1/2) */
	//a  = operator_ho_lowering(qs, 1, param->n_offset);
	//ad = operator_ho_raising(qs, 1, param->n_offset);
	//GSL_SET_COMPLEX(&alpha, param->ho_w[0], 0.0);
	//gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, ad, a, GSL_COMPLEX_ONE, H);
	ws = operator_unit(qs, -1);
	GSL_SET_COMPLEX(&z, 0.5 * param->ho_w[0], 0.0);
	gsl_matrix_complex_scale(ws, z);
	
	n = operator_ho_N(qs, 1, param->n_offset);
	GSL_SET_COMPLEX(&z, param->ho_w[0], 0.0);
	gsl_matrix_complex_scale(n, z);
	gsl_matrix_complex_add(ws, n);
	
	gsl_matrix_complex_add(H, ws);
	gsl_matrix_complex_free(ws);

	/* qubit - cavity interaction */
	a  = operator_ho_lowering(qs, 1, param->n_offset);
	ad = operator_ho_raising(qs, 1, param->n_offset);
	s1 = operator_sigma_x(qs, 0);
	ws = gsl_matrix_complex_alloc(ns, ns);
	gsl_matrix_complex_memcpy(ws, a);
	gsl_matrix_complex_add(ws, ad);
	GSL_SET_COMPLEX(&alpha, param->g_qc, 0.0);
	gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, s1, ws, GSL_COMPLEX_ONE, H);
	gsl_matrix_complex_free(ws);
	gsl_matrix_complex_free(s1);
	
	gsl_matrix_complex_free(a);
	gsl_matrix_complex_free(ad);
	gsl_matrix_complex_free(n);

	return H;
}

/* ---------------------
 * 3-level system + Cavity 
 *
 */

/*
 * Generate the hamiltonian for the 3-level atom + HO + driving field
 *
 */
gsl_matrix_complex *
hamiltonian_3ls_ho_drive(quantum_system *qs, solver_param *param)
{
	int ns;
	gsl_matrix_complex *H, *b, *bd;
	
	ns = quantum_system_nstates(qs);
	
	if ((H = gsl_matrix_complex_alloc(ns, ns)) == NULL)
	{
		fprintf(stderr, "hamiltonian_nTLS: Failed to allocate H. Out of memory?\n");
		return NULL;
	}
	
	b  = operator_ho_lowering(qs, 1, param->n_offset);
	bd = operator_ho_raising(qs, 1, param->n_offset);

	// a + ad
	gsl_matrix_complex_set_zero(H);
	gsl_matrix_complex_add(H, b);
	gsl_matrix_complex_add(H, bd);

	gsl_matrix_complex_free(b);
	gsl_matrix_complex_free(bd);
	
	return H;
}

/*
 * Generate the hamiltonian for the Qubit + Cavity
 *
 */
gsl_matrix_complex *
hamiltonian_3ls_ho(quantum_system *qs, solver_param *param)
{
	int ns;
	gsl_matrix_complex *H, *a, *ad, *ws, *s1, *n;
	gsl_complex z, alpha;
	
	ns = quantum_system_nstates(qs);
	
	if ((H = gsl_matrix_complex_alloc(ns, ns)) == NULL)
	{
		fprintf(stderr, "hamiltonian_nTLS: Failed to allocate H. Out of memory?\n");
		return NULL;
	}

	gsl_matrix_complex_set_zero(H);

	/* qubit */
        s1 = operator_3ls_sigma_z(qs, 0);
	GSL_SET_COMPLEX(&alpha, -0.5 * param->epsilon[0], 0.0);
        gsl_matrix_complex_scale(s1, alpha);
        gsl_matrix_complex_add(H, s1);
        gsl_matrix_complex_free(s1);

        s1 = operator_3ls_sigma_x(qs, 0);
        GSL_SET_COMPLEX(&alpha, -0.5 * param->delta[0], 0.0);
        gsl_matrix_complex_scale(s1, alpha);
        gsl_matrix_complex_add(H, s1);
        gsl_matrix_complex_free(s1);

	/* h_ w1 (ad a + 1/2) */
	//a  = operator_ho_lowering(qs, 1, param->n_offset);
	//ad = operator_ho_raising(qs, 1, param->n_offset);
	//GSL_SET_COMPLEX(&alpha, param->ho_w[0], 0.0);
	//gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, ad, a, GSL_COMPLEX_ONE, H);
	ws = operator_unit(qs, -1);
	GSL_SET_COMPLEX(&z, 0.5 * param->ho_w[0], 0.0);
	gsl_matrix_complex_scale(ws, z);
	
	n = operator_ho_N(qs, 1, param->n_offset);
	GSL_SET_COMPLEX(&z, param->ho_w[0], 0.0);
	gsl_matrix_complex_scale(n, z);
	gsl_matrix_complex_add(ws, n);
	
	gsl_matrix_complex_add(H, ws);
	gsl_matrix_complex_free(ws);

	/* qubit - cavity interaction */
	a  = operator_ho_lowering(qs, 1, param->n_offset);
	ad = operator_ho_raising(qs, 1, param->n_offset);
	s1 = operator_3ls_sigma_z(qs, 0);
	ws = gsl_matrix_complex_alloc(ns, ns);
	gsl_matrix_complex_memcpy(ws, a);
	gsl_matrix_complex_add(ws, ad);
	GSL_SET_COMPLEX(&alpha, param->g_qc, 0.0);
	gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, s1, ws, GSL_COMPLEX_ONE, H);
	gsl_matrix_complex_free(ws);
	gsl_matrix_complex_free(s1);
	
	gsl_matrix_complex_free(a);
	gsl_matrix_complex_free(ad);
	gsl_matrix_complex_free(n);

	return H;
}

/*
 * Generate the hamiltonian for the HO + driving field
 *
 */
gsl_matrix_complex *
hamiltonian_ho_drive(quantum_system *qs, solver_param *param)
{
	int ns;
	gsl_matrix_complex *H, *b, *bd;
	
	ns = quantum_system_nstates(qs);
	
	if ((H = gsl_matrix_complex_alloc(ns, ns)) == NULL)
	{
		fprintf(stderr, "hamiltonian_nTLS: Failed to allocate H. Out of memory?\n");
		return NULL;
	}
	
	b  = operator_ho_lowering(qs, 0, 0);
	bd = operator_ho_raising(qs, 0, 0);

	// a + ad
	gsl_matrix_complex_set_zero(H);
	gsl_matrix_complex_add(H, b);
	gsl_matrix_complex_add(H, bd);
	

	gsl_matrix_complex_free(b);
	gsl_matrix_complex_free(bd);
	
	return H;
}

/*
 * Generate the hamiltonian for the harmonic oscillator
 *
 */
gsl_matrix_complex *
hamiltonian_ho(quantum_system *qs, solver_param *param)
{
	int ns;
	gsl_matrix_complex *H, *ws, *n; //*a, *ad, *ws;
	gsl_complex z;
	
	ns = quantum_system_nstates(qs);
	
	if ((H = gsl_matrix_complex_alloc(ns, ns)) == NULL)
	{
		fprintf(stderr, "hamiltonian_nTLS: Failed to allocate H. Out of memory?\n");
		return NULL;
	}

	gsl_matrix_complex_set_zero(H);

        /* h_ w1 (ad a + 1/2) */
//        a  = operator_ho_lowering(qs, 1, param->n_offset);
//        ad = operator_ho_raising(qs, 1, param->n_offset);
//        GSL_SET_COMPLEX(&alpha, param->ho_w[0], 0.0);
//        gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, ad, a, GSL_COMPLEX_ONE, H);
        ws = operator_unit(qs, -1);
        GSL_SET_COMPLEX(&z, 0.0 * param->ho_w[0], 0.0);
        gsl_matrix_complex_scale(ws, z);

        n = operator_ho_N(qs, 0, param->n_offset);
        GSL_SET_COMPLEX(&z, param->ho_w[0], 0.0);
        gsl_matrix_complex_scale(n, z);
        gsl_matrix_complex_add(ws, n);

        gsl_matrix_complex_add(H, ws);
        gsl_matrix_complex_free(ws);
        gsl_matrix_complex_free(n);

//	/* h_ w1 (ad a + 1/2) */
//	a  = operator_ho_lowering(qs, 0, 0);
//	ad = operator_ho_raising(qs, 0, 0);
//	GSL_SET_COMPLEX(&alpha, param->ho_w[0], 0.0);
//	gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, ad, a, GSL_COMPLEX_ONE, H);
//	ws = operator_unit(qs, -1);
//	GSL_SET_COMPLEX(&z, 0.5 * param->ho_w[0], 0.0);
//	gsl_matrix_complex_scale(ws, z);
//	gsl_matrix_complex_add(H, ws);
//	gsl_matrix_complex_free(ws);
//
//	gsl_matrix_complex_free(a);
//	gsl_matrix_complex_free(ad);

	return H;
}

/* ---------------------------------------------------------------------------------------
 * 3LS + large spin + harmonic oscillator
 *
 */

/*
 * Generate the dicke model hamiltonian + 3LS 
 *
 */
gsl_matrix_complex *
hamiltonian_dicke(quantum_system *qs, solver_param *param)
{
	int ns;
	gsl_matrix_complex *H, *Hsub, *n, *sx, *Jp, *Jm, *J_sum, *a, *ad, *a_sum;
	gsl_complex z;
	
	ns = quantum_system_nstates(qs);
	
	if ((H = gsl_matrix_complex_alloc(ns, ns)) == NULL)
	{
		fprintf(stderr, "hamiltonian_nTLS: Failed to allocate H. Out of memory?\n");
		return NULL;
	}

	/* 
 	 * start by blanking the hamiltonian matrix
 	 */
	gsl_matrix_complex_set_zero(H);

	/*
 	 * subsystem 1: 3LS artificial atom: 
 	 *
 	 * /
        Hsub = operator_sigma_z(qs, 0);
        GSL_SET_COMPLEX(&z, 0.5 * param->epsilon[0], 0.0);
        gsl_matrix_complex_scale(Hsub, z);
        gsl_matrix_complex_add(H, Hsub);
        gsl_matrix_complex_free(Hsub);

        Hsub = operator_sigma_x(qs, 0);
        GSL_SET_COMPLEX(&z, 0.5 * param->delta[0], 0.0);
        gsl_matrix_complex_scale(Hsub, z);
        gsl_matrix_complex_add(H, Hsub);
        gsl_matrix_complex_free(Hsub);
	*/

	/* 
 	 * subsystem 2: N/2 spin:
 	 *
 	 */
	Hsub = operator_Jz(qs, 1-1, param->j);
        GSL_SET_COMPLEX(&z, param->J_w[0], 0.0);
        gsl_matrix_complex_scale(Hsub, z);

	gsl_matrix_complex_add(H, Hsub);
	gsl_matrix_complex_free(Hsub);

	/* 
	 * subsystem 3: HO: h_ w1 (ad a + 1/2) 
	 *
	 */
        Hsub = operator_ho_N(qs, 2-1, param->n_offset);
        GSL_SET_COMPLEX(&z, -param->ho_w[0], 0.0);
        gsl_matrix_complex_scale(Hsub, z);

        gsl_matrix_complex_add(H, Hsub);
        gsl_matrix_complex_free(Hsub);

	/*
 	 * Interaction: 
 	 * 1) artificial atom <-> oscillator
 	 * 2) spin <-> oscillator
 	 *
 	 */
        a  = operator_ho_lowering(qs, 2-1, param->n_offset);
        ad = operator_ho_raising (qs, 2-1, param->n_offset);
        a_sum = gsl_matrix_complex_alloc(ns, ns);
        gsl_matrix_complex_memcpy(a_sum, a);
        gsl_matrix_complex_add   (a_sum, ad);
        
	// 1)
	/*
	sx = operator_sigma_x(qs, 0);

        GSL_SET_COMPLEX(&z, param->g_ac, 0.0);
        gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, z, sx, a_sum, GSL_COMPLEX_ONE, H);

        gsl_matrix_complex_free(sx);
	*/
	// 2)
	
	Jp = operator_J_plus (qs, 1-1, param->j);
	Jm = operator_J_minus(qs, 1-1, param->j);
	J_sum = gsl_matrix_complex_alloc(ns, ns);
        gsl_matrix_complex_memcpy(J_sum, Jp);
        gsl_matrix_complex_add   (J_sum, Jm);

        GSL_SET_COMPLEX(&z, param->g_sc, 0.0);
        gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, z, J_sum, a_sum, GSL_COMPLEX_ONE, H);

        gsl_matrix_complex_free(Jp);
        gsl_matrix_complex_free(Jm);
        gsl_matrix_complex_free(J_sum);

        gsl_matrix_complex_free(a);
        gsl_matrix_complex_free(ad);
        gsl_matrix_complex_free(a_sum);

	return H;
}





