/*
 * Functions for integrating systems of differential equations. 
 * 
 * Robert Johansson <robert@riken.jp>
 *
 * $Id: integrate.c,v 1.1.1.1 2008-09-03 09:36:41 rob Exp $
 */

#include <math.h>
#include <stdio.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_complex.h>
#include <gsl/gsl_complex_math.h>

#include "quantum_system.h"
#include "integrate.h"
#include "param.h"


/**
 * \breif	Evolve a density matrix EQM using a standard Runge-Kutta.
 *
 * Standard Runge-Kutta method.
 *
 */
int 
odeint_complex_rk(quantum_system *qs, solver_param *p, gsl_matrix_complex *rho0,
		  ode_func_drho_dt f_drho_dt, rho_cb_func_t rho_cb_func)
{
	gsl_matrix_complex *rho_t, *ws, *k1, *k2, *k3, *k4;
	gsl_complex z;
	double t;
	int ns;

	ns = quantum_system_nstates(qs);

	rho_t  = gsl_matrix_complex_alloc(ns, ns);
	ws     = gsl_matrix_complex_alloc(ns, ns);
	k1     = gsl_matrix_complex_alloc(ns, ns);
	k2     = gsl_matrix_complex_alloc(ns, ns);
	k3     = gsl_matrix_complex_alloc(ns, ns);
	k4     = gsl_matrix_complex_alloc(ns, ns);

	gsl_matrix_complex_memcpy(rho_t, rho0);
	
	rho_cb_func(rho_t, p->Ti, qs, p);
	
	for (t = p->Ti; t < p->Tf; t += p->dT)
	{
		gsl_matrix_complex_memcpy(ws, rho_t);
		f_drho_dt(ws, k1, t, qs, p);
		GSL_SET_COMPLEX(&z, p->dT, 0);
		gsl_matrix_complex_scale(k1, z);
		
		gsl_matrix_complex_memcpy(ws, k1);
		GSL_SET_COMPLEX(&z, 0.5, 0);
		gsl_matrix_complex_scale(ws, z);
		gsl_matrix_complex_add(ws, rho_t);
		f_drho_dt(ws, k2, t + p->dT / 2.0, qs, p);
		GSL_SET_COMPLEX(&z, p->dT, 0);
		gsl_matrix_complex_scale(k2, z);
		
		gsl_matrix_complex_memcpy(ws, k2);
		GSL_SET_COMPLEX(&z, 0.5, 0);
		gsl_matrix_complex_scale(ws, z);
		gsl_matrix_complex_add(ws, rho_t);
		f_drho_dt(ws, k3, t + p->dT / 2.0, qs, p);
		GSL_SET_COMPLEX(&z, p->dT, 0);
		gsl_matrix_complex_scale(k3, z);
		
		gsl_matrix_complex_memcpy(ws, k3);
		gsl_matrix_complex_add(ws, rho_t);
		f_drho_dt(ws, k4, t + p->dT, qs, p);
		GSL_SET_COMPLEX(&z, p->dT, 0);
		gsl_matrix_complex_scale(k4, z);
		
		GSL_SET_COMPLEX(&z, 2.0, 0);
		gsl_matrix_complex_memcpy(ws, k1);
		gsl_matrix_complex_scale(k2, z);
		gsl_matrix_complex_add(ws, k2);	
		gsl_matrix_complex_scale(k3, z);
		gsl_matrix_complex_add(ws, k3);	
		gsl_matrix_complex_add(ws, k4);	
		GSL_SET_COMPLEX(&z, 1/6.0, 0);
		gsl_matrix_complex_scale(ws, z);
		
		// rho_t(t+dt) = rho(t) + 1/6 * (k1 + 2 k2 + 2 k3 + k4)
		gsl_matrix_complex_add(rho_t, ws);	
		rho_cb_func(rho_t, t, qs, p);
	}

	return 0;
}


/*
 * Two-step adaptive Runge-Kutta method with local error estimate
 * and step-length control.
 */
int 
odeint_complex_rk_adaptive(quantum_system *qs, solver_param *p, gsl_matrix_complex *rho0,
		           ode_func_drho_dt f_drho_dt, rho_cb_func_t rho_cb_func)
{
	gsl_matrix_complex *rho1, *rho2;
	double t, e, emax = 0.0, etol = 0.001;
	int ns;

	ns = quantum_system_nstates(qs);

	rho1  = gsl_matrix_complex_alloc(ns, ns);
	rho2  = gsl_matrix_complex_alloc(ns, ns);
	gsl_matrix_complex_memcpy(rho1, rho0);
	
	rho_cb_func(rho1, p->Ti, qs, p);
	
	for (t = p->Ti; t < p->Tf - p->dT; t += p->dT)
	{
		int rd;
		
		rd = odeint_rk_adaptive_rec(qs, p, f_drho_dt, rho1, rho2, 0, t, t+p->dT, etol, &e);
		if (e > emax)
			emax = e;
		
		//printf("odeint_complex_rk_adaptive: Local error estimate %f (rec depth %d) at time t = %f\n", e, rd, t);
		
		gsl_matrix_complex_memcpy(rho1, rho2);
		rho_cb_func(rho1, t + p->dT, qs, p);
	}
	
	//printf("odeint_complex_rk_adaptive: Maximum (estimated) local error throughout simulation: %f\n", emax);
	
	return 0;
}

/*
 * Recursive method for calculating a two-step Runge-Kutta on shrinking interval.
 *
 */
int 
odeint_rk_adaptive_rec(quantum_system *qs, solver_param *p, ode_func_drho_dt f_drho_dt,
                           gsl_matrix_complex *rho0, gsl_matrix_complex *rho_final,
                           int level, double ti, double tf, double etol, double *e)
{
	double e1, e2, le, h;
	gsl_complex hz, z;
	gsl_matrix_complex *rhot, *rho1, *rho2, *ws, *k1, *k2, *k3, *k4;
	unsigned int i, j; 
	int ns, rd = level;

	ns = quantum_system_nstates(qs);

	rhot = gsl_matrix_complex_alloc(ns, ns);
	rho1 = gsl_matrix_complex_alloc(ns, ns);
	rho2 = gsl_matrix_complex_alloc(ns, ns);
	ws   = gsl_matrix_complex_alloc(ns, ns);
	k1   = gsl_matrix_complex_alloc(ns, ns); // can be optimized by using k and kacc
	k2   = gsl_matrix_complex_alloc(ns, ns);
	k3   = gsl_matrix_complex_alloc(ns, ns);
	k4   = gsl_matrix_complex_alloc(ns, ns);

	*e = 0.0;
	
	/*
	 * calculate rho1 using a one-step RK 
	 * 
	 * rho1 = rho0 + 1/6 (k1 + 2 k2 + 2 k3 + k4)
	 * alt: copy rho2 to rho1, if level != 0
	 */
	h = tf - ti;
	GSL_SET_COMPLEX(&hz, h, 0.0);
	GSL_SET_COMPLEX(&z, 0.5, 0);
	
	// K1
	gsl_matrix_complex_memcpy(ws, rho0);
	f_drho_dt(ws, k1, ti, qs, p);
	gsl_matrix_complex_scale(k1, hz);
	
	// K2
	gsl_matrix_complex_memcpy(ws, k1);
	gsl_matrix_complex_scale(ws, z);
	gsl_matrix_complex_add(ws, rho0);
	f_drho_dt(ws, k2, ti + h / 2.0, qs, p);
	gsl_matrix_complex_scale(k2, hz);
	
	// K3
	gsl_matrix_complex_memcpy(ws, k2);
	gsl_matrix_complex_scale(ws, z);
	gsl_matrix_complex_add(ws, rho0);
	f_drho_dt(ws, k3, ti + h / 2.0, qs, p);
	gsl_matrix_complex_scale(k3, hz);
	
	// K4
	gsl_matrix_complex_memcpy(ws, k3);
	gsl_matrix_complex_add(ws, rho0);
	f_drho_dt(ws, k4, ti + h, qs, p);
	gsl_matrix_complex_scale(k4, hz);
		
	GSL_SET_COMPLEX(&z, 2.0, 0);
	gsl_matrix_complex_memcpy(ws, k1);
	gsl_matrix_complex_scale(k2, z);
	gsl_matrix_complex_add(ws, k2);	
	gsl_matrix_complex_scale(k3, z);
	gsl_matrix_complex_add(ws, k3);	
	gsl_matrix_complex_add(ws, k4);	
	GSL_SET_COMPLEX(&z, 1/6.0, 0);
	gsl_matrix_complex_scale(ws, z);
		
	// rho_t(t+dt) = rho(t) + 1/6 * (k1 + 2 k2 + 2 k3 + k4)
	gsl_matrix_complex_memcpy(rho1, rho0);
	gsl_matrix_complex_add(rho1, ws);	

	/* 
	 * calculate rho2 using a two-step RK 
	 *
	 * 
	 */
	
	h = (tf - ti) / 2.0;
	GSL_SET_COMPLEX(&hz, h, 0.0);
	
	/* --- STEP 1 --- */
	GSL_SET_COMPLEX(&z, 0.5, 0);
	// K1
	gsl_matrix_complex_memcpy(ws, rho0);
	f_drho_dt(ws, k1, ti, qs, p);
	gsl_matrix_complex_scale(k1, hz);
	
	// K2
	gsl_matrix_complex_memcpy(ws, k1);
	gsl_matrix_complex_scale(ws, z);
	gsl_matrix_complex_add(ws, rho0);
	f_drho_dt(ws, k2, ti + h / 2.0, qs, p);
	gsl_matrix_complex_scale(k2, hz);
	
	// K3
	gsl_matrix_complex_memcpy(ws, k2);
	gsl_matrix_complex_scale(ws, z);
	gsl_matrix_complex_add(ws, rho0);
	f_drho_dt(ws, k3, ti + h / 2.0, qs, p);
	gsl_matrix_complex_scale(k3, hz);
	
	// K4
	gsl_matrix_complex_memcpy(ws, k3);
	gsl_matrix_complex_add(ws, rho0);
	f_drho_dt(ws, k4, ti + h, qs, p);
	gsl_matrix_complex_scale(k4, hz);
		
	GSL_SET_COMPLEX(&z, 2.0, 0);
	gsl_matrix_complex_memcpy(ws, k1);
	gsl_matrix_complex_scale(k2, z);
	gsl_matrix_complex_add(ws, k2);	
	gsl_matrix_complex_scale(k3, z);
	gsl_matrix_complex_add(ws, k3);	
	gsl_matrix_complex_add(ws, k4);	
	GSL_SET_COMPLEX(&z, 1/6.0, 0);
	gsl_matrix_complex_scale(ws, z);
		
	// rhot = rho0 + 1/6 * (k1 + 2 k2 + 2 k3 + k4)
	gsl_matrix_complex_memcpy(rhot, rho0);
	gsl_matrix_complex_add(rhot, ws);	
	
	/* --- STEP 2 --- */
	GSL_SET_COMPLEX(&z, 0.5, 0);
	// K1
	gsl_matrix_complex_memcpy(ws, rhot);
	f_drho_dt(ws, k1, ti + h, qs, p);
	gsl_matrix_complex_scale(k1, hz);
	
	// K2
	gsl_matrix_complex_memcpy(ws, k1);
	gsl_matrix_complex_scale(ws, z);
	gsl_matrix_complex_add(ws, rhot);
	f_drho_dt(ws, k2, ti + 3 * h / 2.0, qs, p);
	gsl_matrix_complex_scale(k2, hz);
	
	// K3
	gsl_matrix_complex_memcpy(ws, k2);
	gsl_matrix_complex_scale(ws, z);
	gsl_matrix_complex_add(ws, rhot);
	f_drho_dt(ws, k3, ti + 3 * h / 2.0, qs, p);
	gsl_matrix_complex_scale(k3, hz);
	
	// K4
	gsl_matrix_complex_memcpy(ws, k3);
	gsl_matrix_complex_add(ws, rhot);
	f_drho_dt(ws, k4, ti + 2 * h, qs, p);
	gsl_matrix_complex_scale(k4, hz);
		
	GSL_SET_COMPLEX(&z, 2.0, 0);
	gsl_matrix_complex_memcpy(ws, k1);
	gsl_matrix_complex_scale(k2, z);
	gsl_matrix_complex_add(ws, k2);	
	gsl_matrix_complex_scale(k3, z);
	gsl_matrix_complex_add(ws, k3);	
	gsl_matrix_complex_add(ws, k4);	
	GSL_SET_COMPLEX(&z, 1/6.0, 0);
	gsl_matrix_complex_scale(ws, z);
		
	// rho2 = rhot + 1/6 * (k1 + 2 k2 + 2 k3 + k4)
	gsl_matrix_complex_memcpy(rho2, rhot);
	gsl_matrix_complex_add(rho2, ws);	

	/* 
	 * Combine rho1 and rho2 into a new estimate rho 
	 *
	 */
	gsl_matrix_complex_memcpy(rho_final, rho2);
	GSL_SET_COMPLEX(&z, 16.0, 0.0);
	gsl_matrix_complex_scale(rho_final, z);
	gsl_matrix_complex_sub(rho_final, rho1);
	GSL_SET_COMPLEX(&z, 1/15.0, 0.0);
	gsl_matrix_complex_scale(rho_final, z);

	/* 
	 * Find maximum deviation 
	 *
	 */
	for (i = 0; i < rho_final->size1; i++)	
	{
		for (j = 0; j < rho_final->size1; j++)
		{
			gsl_complex r_ij, r1_ij, r2_ij;
			
			r_ij  = gsl_matrix_complex_get(rho_final,i,j);
		
			if (gsl_complex_abs(r_ij) != 0.0)
			{
				r1_ij = gsl_matrix_complex_get(rho1,i,j);
				r2_ij = gsl_matrix_complex_get(rho2,i,j);
				
				z = gsl_complex_sub(r1_ij, r2_ij);
				
				le = (gsl_complex_abs(z)) / gsl_complex_abs(r_ij);
				
				if (le > *e && gsl_complex_abs(r_ij) > etol/10.0)
				{
					*e = le;
				}
			}
		}
	}
	
//	printf("DEBUG: odeint_rk_adaptive_rec: local error e = %f,\t[ti, tf] = [%f, %f], level = %d\n", *e, ti, tf, level);

	/*
	 * Recursively refine intervals if errors are too large
	 *
	 */
	if (*e > etol && level < RK_REC_MAX_LEVEL)
	{
		int rd1, rd2;
		rd1 = odeint_rk_adaptive_rec(qs, p, f_drho_dt, rho0, rhot, level + 1, ti, ti+(tf-ti)/2.0, etol, &e1);
		rd2 = odeint_rk_adaptive_rec(qs, p, f_drho_dt, rhot, rho_final, level + 1, ti+(tf-ti)/2.0, tf, etol, &e2);
		*e = (e1 > e2) ? e1 : e2;
		rd = (rd1 > rd2) ? rd1 : rd2;
	}
	else
	{
		rd = level;
	}

	gsl_matrix_complex_free(rhot);
	gsl_matrix_complex_free(rho1);
	gsl_matrix_complex_free(rho2);
	gsl_matrix_complex_free(ws);
	gsl_matrix_complex_free(k1);
	gsl_matrix_complex_free(k2);
	gsl_matrix_complex_free(k3);
	gsl_matrix_complex_free(k4);
	
	return rd;
}


/* 
 * Find maximum deviation 
 *
 */
double
max_deviation(gsl_matrix_complex *m1, gsl_matrix_complex *m2, double etol)
{
	unsigned int i, j;
	double emax = 0.0, le;
	
	for (i = 0; i < m1->size1; i++)	
	{
		for (j = 0; j < m2->size1; j++)
		{
			gsl_complex m1_ij, m2_ij;
			
			m2_ij  = gsl_matrix_complex_get(m2,i,j);
		
			if (gsl_complex_abs(m2_ij) != 0.0)
			{
				m1_ij = gsl_matrix_complex_get(m1,i,j);
			
				le = gsl_complex_abs(m1_ij) / gsl_complex_abs(m2_ij);
			
				if (le > emax && gsl_complex_abs(m2_ij) > etol/10.0)
				{
					emax = le;
				}
			}
		}
	}

	return emax;
}
