/*
 * A collection of various functions: printing, saving files, etc. 
 * 
 * 
 * Robert Johansson <robert@riken.jp>
 *
 * $Id: misc.c,v 1.1.1.1 2008-09-03 09:36:41 rob Exp $
 */
#include <gsl/gsl_matrix.h>

#define _GNU_SOURCE
#include <stdio.h>


#include "misc.h"

/*
 *
 *
 */
void
misc_matrix_print(gsl_matrix *m, int n)
{
	int i, j;
	
	for (i = 0; i < n; i++) 
	{
   		for (j = 0; j < n; j++)
	        	printf ("%2.5f  ", gsl_matrix_get(m, i, j));
		printf("\n");
	}
}

/*
 * Print complex matrix with element format "(real, imag)"
 *
 */
void
misc_matrix_complex_print(gsl_matrix_complex *m, int n)
{
	double val_re, val_im;
	int i, j;
	
	for (i = 0; i < n; i++) 
	{
   		for (j = 0; j < n; j++)
		{	
			val_re = GSL_REAL(gsl_matrix_complex_get(m, i, j));
			val_im = GSL_IMAG(gsl_matrix_complex_get(m, i, j));
	        	printf (" (%2.5f,%2.5f)", val_re, val_im);
		}
		printf("\n");
	}
}
/*
 *
 *
 */
void
misc_matrix_complex_print_real(gsl_matrix_complex *m, int n)
{
	double val;
	int i, j;
	
	for (i = 0; i < n; i++) 
	{
   		for (j = 0; j < n; j++)
		{	
			val = GSL_REAL(gsl_matrix_complex_get(m, i, j));
			if (val >= 0.0)
	        		printf (" %2.5f  ", val);
			else
	        		printf ("%2.5f  ", val);
		}
		printf("\n");
	}
}

/*
 *
 *
 */
void
misc_matrix_complex_print_imag(gsl_matrix_complex *m, int n)
{
	int i, j;
	
	for (i = 0; i < n; i++) 
	{
   		for (j = 0; j < n; j++)
	        	printf ("%2.5f  ", GSL_IMAG(gsl_matrix_complex_get(m, i, j)));
		printf("\n");
	}
}


/*
 *
 *
 *
 */
int
misc_file_row_add(const char *filename, const char *row)
{
	FILE *file;

//	while ((file = fopen(filename, "ax")) == NULL)
//	{
//		if (errno == EEXIST)
//			continue;
//		
//		fprintf(stderr, "misc_file_row_add: Failed to open file %s\n", filename);
//		return -1;
//	}


	if ((file = fopen(filename, "a")) == NULL)
	{
		fprintf(stderr, "misc_file_row_add: Failed to open file %s\n", filename);
		return -1;
	}

	fprintf(file, "%s\n", row);
	
	fclose(file);

	return 0;
}

/*
 * Write/read matrix to/from file:
 *
 * Format
 *
 * a11 a12 a13 ... a1N 
 * a21 a22 a23 ... a2N 
 * ... 
 * aN1 aN2 aN3 ... aNN
 *
 * Arguments:
 *
 *  m         = matrix to write to
 *  filename  = name of file to read from
 *  n         = dimension of (square) matrix
 *
 */
int
misc_matrix_write(char *filename, gsl_matrix_complex *m, int n)
{
	int i, j;
        FILE *file;

        if ((file = fopen(filename, "w")) == NULL)
        {
                fprintf(stderr, "misc_matrix_write: Failed to open file %s\n", filename);
                return -1;
        }

	for (i = 0; i < n; i++)
	{
		for (j = 0; j < n; j++)
		{
			gsl_complex z;
			z = gsl_matrix_complex_get(m, i, j);
			fprintf(file, "(%f, %f) ", GSL_REAL(z), GSL_IMAG(z));
		}
		fprintf(file, "\n");
	}

        fclose(file);

        return 0;
}

int
misc_matrix_read(char *filename, gsl_matrix_complex *m, int n)
{
	int i = 0, j;
	char *linebuff = NULL; //[100000];
        FILE *file;
	size_t size;

        if ((file = fopen(filename, "r")) == NULL)
        {
                fprintf(stderr, "misc_matrix_read: Failed to open file %s\n", filename);
                return -1;
        }

	while (getline(&linebuff, &size, file) > 0)
	{
		gsl_complex z;
		double zre, zim;
		char *endptr, *ptr;
		j = 0;
		

		ptr = &linebuff[1];

		while (j < n)
		{
			zre = strtod(ptr, &endptr);
			ptr = endptr + 1;
			zim = strtod(ptr, &endptr);
			ptr = endptr + 3;

			GSL_SET_COMPLEX(&z, zre, zim);
			
			gsl_matrix_complex_set(m, i, j, z);

			j++;
		}
		i++;

		//free(linebuff);
	}

	return 0;	
}



/*
 * Write/read diagonal matrix to/from file:
 *
 * Format
 *
 * a11 a22 a33 ... aNN 
 *
 * Arguments:
 *
 *  m         = matrix to write to
 *  filename  = name of file to read from
 *  n         = dimension of (square) matrix
 *
 */
int
misc_matrix_diag_write(char *filename, gsl_matrix_complex *m, int n, char *msg)
{
	int i;
        FILE *file;

        if ((file = fopen(filename, "w")) == NULL)
        {
                fprintf(stderr, "misc_matrix_read_write: Failed to open file %s\n", filename);
                return -1;
        }

	fprintf(file, "%s", msg);
	for (i = 0; i < n; i++)
	{
		gsl_complex z;
		z = gsl_matrix_complex_get(m, i, i);
		fprintf(file, "(%f, %f) ", GSL_REAL(z), GSL_IMAG(z));
	}
	fprintf(file, "\n");

        fclose(file);
        return 0;
}

int
misc_matrix_diag_read(char *filename, gsl_matrix_complex *m, int n)
{
	int i = 0, j;
	char *linebuff = NULL;
        FILE *file;
	size_t size;

        if ((file = fopen(filename, "r")) == NULL)
        {
                fprintf(stderr, "misc_matrix_diag_read: Failed to open file %s\n", filename);
                return -1;
        }

	gsl_matrix_complex_set_zero(m);

	if (getline(&linebuff, &size, file) > 0)
	{
		gsl_complex z;
		double zre, zim;
		char *endptr, *ptr;
		j = 0;

		ptr = &linebuff[1];

		while (j < n)
		{
			zre = strtod(ptr, &endptr);
			ptr = endptr + 1;
			zim = strtod(ptr, &endptr);
			ptr = endptr + 3;

			GSL_SET_COMPLEX(&z, zre, zim);
			
			gsl_matrix_complex_set(m, j, j, z);

			j++;
		}
		//free(linebuff);
	}

	return 0;	
}

int
misc_matrix_diag_real_write(char *filename, gsl_matrix_complex *m, int n, char *msg)
{
	int i;
        FILE *file;

        if ((file = fopen(filename, "w")) == NULL)
        {
                fprintf(stderr, "misc_matrix_diag_real_write: Failed to open file %s\n", filename);
                return -1;
        }

	fprintf(file, "%s", msg);
	for (i = 0; i < n; i++)
	{
		gsl_complex z;
		z = gsl_matrix_complex_get(m, i, i);
		fprintf(file, "%f ", GSL_REAL(z));
	}
	fprintf(file, "\n");

        fclose(file);
        return 0;
}


int
misc_matrix_diag_real_read(char *filename, gsl_matrix_complex *m, int n)
{
	int i = 0, j;
	char *linebuff = NULL;
        FILE *file;
	size_t size;

        if ((file = fopen(filename, "r")) == NULL)
        {
                fprintf(stderr, "misc_matrix_diag_real_read: Failed to open file %s\n", filename);
                return -1;
        }

	gsl_matrix_complex_set_zero(m);

	if (getline(&linebuff, &size, file) > 0)
	{
		gsl_complex z;
		double zre, zim;
		char *endptr, *ptr;
		j = 0;

		ptr = linebuff;

		while (j < n)
		{
			zre = strtod(ptr, &endptr);
			ptr = endptr;

			GSL_SET_COMPLEX(&z, zre, 0.0);
			
			gsl_matrix_complex_set(m, j, j, z);

			j++;
		}
		//free(linebuff);
	}

	return 0;	
}









