/*
 * Functions related to definition of quantum systems (composed of subsystems
 * with different sizes) and enummeration of states.
 *
 *
 * Robert Johansson <robert@riken.jp>
 *
 * $Id: quantum_system.c,v 1.1.1.1 2008-09-03 09:36:41 rob Exp $
 */
#include "quantum_system.h"

#include <stdlib.h>
#include <stdio.h>

/** 
 * Create a new quantum_system object.
 *
 */
quantum_system *
quantum_system_new()
{
	quantum_system *qs;

	if ((qs = malloc(sizeof(quantum_system))) == NULL)
	{
		fprintf(stderr, "Failed ot allocate a new quantum_system object\n");
		return 0;
	}

	qs->nsubsys = 0;
	qs->nstates = NULL;

	return qs;
}

/*
 * Create a new quantum_system object using qs as a template.
 *
 *
 */
quantum_system *
quantum_system_copy(quantum_system *qs)
{
	int i;
	quantum_system *qs_new;

	qs_new = quantum_system_new();
	for (i = 0; i < qs->nsubsys; i++)
		quantum_system_add(qs_new, qs->nstates[i]);
	
	return qs_new;
}


/*
 *
 *
 *
 */
void
quantum_system_add(quantum_system *qs, int M)
{
	int *nss, i;

	nss = malloc(sizeof(int) * (qs->nsubsys + 1));

	for (i = 0; i < qs->nsubsys; i++)
		nss[i] = qs->nstates[i];
	nss[qs->nsubsys] = M;
	qs->nsubsys = qs->nsubsys + 1;
	free(qs->nstates);
	qs->nstates = nss;
}

/*
 *
 *
 */
void
quantum_system_print(quantum_system *qs)
{
	int i;

	printf("Quantum system with %d sub systems: ", qs->nsubsys);
	for (i = 0; i < qs->nsubsys; i++)
		printf("%d ", qs->nstates[i]);
	printf("\n");
}

/*
 *
 *
 */
void
quantum_system_free(quantum_system *qs)
{
	free(qs->nstates);
	free(qs);
}



/*
 * Calculate the number of quantum states in the full system.
 */
int
quantum_system_nstates(quantum_system *qs)
{
	int i, ns = 1;

	for (i = 0; i < qs->nsubsys; i++)
	{
		ns *= qs->nstates[i];
	}

	return ns;
}

/*
 * Calculate the quantum state vector that correspond to the quantum 
 * state number qsn.
 *
 */
void
quantum_system_state_vector(quantum_system *qs, int qsn, quantum_system_state_vector_t qsv)
{
	int i, acc;

	acc = quantum_system_nstates(qs);

	for (i = qs->nsubsys-1; i >= 0; i--)
	{
		acc = acc / qs->nstates[i];
		qsv[i] = floor(qsn/acc);
		qsn -= qsv[i] * acc;
	}
}

/*
 * State vectors
 * 
 */
int
quantum_system_state_vector_print(quantum_system *qs, quantum_system_state_vector_t qsv)
{
	int i;

	for (i = 0; i < qs->nsubsys; i++)
		printf("%d ", qsv[i]);

	return 0;
}

/*
 * Calculate the quantum state number that corresponds to the quantum 
 * state vector qsv.
 */
int
quantum_system_state_number(quantum_system *qs, quantum_system_state_vector_t qsv)
{
	int qsn, i, acc_nqs;
	
	qsn = 0;
	acc_nqs = 1;

	for (i = 0; i < qs->nsubsys; i++)
	{
		qsn += acc_nqs * qsv[i];
		acc_nqs *= qs->nstates[i];
	}

	return qsn;
}




		
