/*
 * Find the steady state of a quantum system with dissipation. 
 * 
 * Robert Johansson <robert@riken.jp>
 *
 * $Id: steadystate.c,v 1.3 2008-11-04 02:49:32 rob Exp $
 */

#include <stdio.h>
#include <gsl/gsl_complex.h>
#include <gsl/gsl_complex_math.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_linalg.h>

#include "quantum_system.h"
#include "operators.h"
#include "hamiltonian.h"
#include "master_equation.h"
#include "param.h"
#include "misc.h"
#include "integrate.h"
#include "gsl_ext.h"
#include "gsl_ext_expm_complex.h"
#include "basis_transform.h"

/* ----------------------------------------------------------
 * Helper functions:
 *
 */
//#define KRON(a,b)	(a==b)


/* ----------------------------------------------------------
 * Index conversion functions:
 *
 */

/*
 * Calculate the superoperator index given two operator indices.
 * 
 * J = N*a+b.
 */
static int
ss_idx_major(int N, int a, int b)
{
//	printf("idx_major(%d, %d)[%d] = %d\n", a, b, N, N*a+b);
	return N*a + b;
}

/*
 * Calculate the first (most significant: ms) operator index 
 * given one super-operator index.
 *
 * J = N*a + b,
 * Solve for a.
 */
static int
ss_idx_minor_ms(int N, int J)
{
	return floor(((double)J)/N);
}

/*
 * Calculate the second (least significant: ls) operator index 
 * given one super-operator index.
 *
 * J = N*a + b,
 * Solve for b.
 */
static int
ss_idx_minor_ls(int N, int J)
{
	return J - N * floor(((double)J)/N);
}


/* ----------------------------------------------------------
 * Functions for calculating the Liouvillian.
 *
 */
static int
steady_state_L_add_op(gsl_matrix_complex *sop_L,
 		      gsl_matrix_complex *op_left,
		      gsl_matrix_complex *op_right,
		      gsl_complex z, int N)
{
	gsl_complex alpha, beta, gamma;

	int a, b, c, d, I, J;

	printf("DEBUG: steady_state_L_add_op: starting...");

	if (op_left != NULL && op_right == NULL)
	{
		/* 
		 * Left operators:
		 *
		 * L_{ab}{cd} = O_{ac}KRON{b,d} 
		 *
		 * ->
		 *
		 * non-zero elements at L_{ab}{cb} = O_{ac}
	         *
		 */
		for (c = 0; c < N; c++)
		{
			for (a = 0; a < N; a++)
			{
				// alpha = op_left(a,c) * z
                                alpha = gsl_matrix_complex_get(op_left, a, c);
                                alpha = gsl_complex_mul(alpha, z);

				if (GSL_REAL(alpha) != 0.0 || GSL_IMAG(alpha) != 0.0)
				{
					for (b = 0; b < N; b++)
					{
						I = ss_idx_major(N, a, b);
						J = ss_idx_major(N, c, b);
	
						// sop_L(I,J) = sop_L(I,J) + alpha
						beta  = gsl_matrix_complex_get(sop_L, I, J);
						gamma = gsl_complex_add(alpha, beta);
						gsl_matrix_complex_set(sop_L, I, J, gamma);
					}
				}
			}
		}
	}
	
	
	/* 
	 * right operators:
	 *
	 * L_{ab}{cd} = O_{bd}KRON{ca} 
	 *
	 * ->
	 *
	 * non-zero elements at L_{ab}{ad} = O_{bd}
         *
	 */
	if (op_left == NULL && op_right != NULL)
	{
		for (d = 0; d < N; d++)
		{
			for (b = 0; b < N; b++)
			{
				// alpha = op_right(b,d) * z
				alpha = gsl_matrix_complex_get(op_right, b, d);
				alpha = gsl_complex_mul(alpha, z);
				
				if (GSL_REAL(alpha) != 0.0 || GSL_IMAG(alpha) != 0.0)
				{
					for (a = 0; a < N; a++)
					{
						I = ss_idx_major(N, a, b);
						J = ss_idx_major(N, a, d);
	
						// sop_L(I,J) = sop_L(I,J) + alpha
						beta  = gsl_matrix_complex_get(sop_L, I, J);
						gamma = gsl_complex_add(alpha, beta);
						gsl_matrix_complex_set(sop_L, I, J, gamma);
					}
				}
			}
		}
	}
	
	
	/* 
	 * left and right operators simultaneously:
	 *
	 * L_{ab}{cd} = A_{ac}B_{db}
	 *
	 */
	if (op_left != NULL && op_right != NULL)
	{
		for (a = 0; a < N; a++)
		{
			for (c = 0; c < N; c++)
			{
				// alpha = op_left(a,c) * op_right(d,b) * z
                                alpha = gsl_matrix_complex_get(op_left,  a, c);
				alpha = gsl_complex_mul(alpha, z);
				
				if (GSL_REAL(alpha) != 0.0 || GSL_IMAG(alpha) != 0.0)
				{
					for (b = 0; b < N; b++)
					{
						for (d = 0; d < N; d++)
						{
							I = ss_idx_major(N, a, b);
							J = ss_idx_major(N, c, d);

							beta  = gsl_matrix_complex_get(op_right, d, b);

							if (GSL_REAL(beta) != 0.0 || GSL_IMAG(beta) != 0.0)
                                			{
								gamma = gsl_complex_mul(alpha, beta);

                	        	                	// sop_L(I,J) = sop_L(I,J) + alpha
                	        	                	beta  = gsl_matrix_complex_get(sop_L, I, J);
                	        	                	gamma = gsl_complex_add(gamma, beta);
                	       		                	gsl_matrix_complex_set(sop_L, I, J, alpha);
							}
						}
					}
				}
			}
		}
	}
	
	printf("done.\n");
	
	return 0;
}

/*
 * Find the steady-state density matrix for a quantum system with dissipation by
 * calculate the system Liouville super operators, and solving L rho = 0, including
 * the density normalisation condition.
 *
 *
 */
int
dm_steady_state(quantum_system *qs, gsl_matrix_complex *rho_ss, solver_param *param)
{
	gsl_matrix_complex *sop_L;
	gsl_vector_complex *rho_ss_vec, *ss_eq_rhs;
	gsl_permutation *p;

	gsl_complex z, alpha, beta;
	int ns, s, i;

	ns = quantum_system_nstates(qs);

	printf("DEBUG: solving for rho_ss: starting\n");

	/*
 	 * Allocate space for the L super-operator and
  	 * some working space vectors for solving the
	 * LA problem.
 	 */
	sop_L 	   = gsl_matrix_complex_alloc(ns*ns, ns*ns);
	rho_ss_vec = gsl_vector_complex_alloc(ns*ns);
	ss_eq_rhs  = gsl_vector_complex_alloc(ns*ns);
	p 	   = gsl_permutation_alloc(ns*ns);

	gsl_matrix_complex_set_zero(sop_L);

        /* --- Precalculate lindblad operators --- */
	param->H0 = param->H0_func(qs, param);

        for (i = 0; i < param->do_n; i++)
        {
                GSL_SET_COMPLEX(&alpha, 1.0, 0.0);
                GSL_SET_COMPLEX(&beta,  0.0, 0.0);

                param->do_aad[i] = NULL;
                param->do_ada[i] = NULL;

                if (param->do_g2[i] != 0.0)
                {
                        param->do_aad[i] = gsl_matrix_complex_alloc(ns, ns);
                        gsl_blas_zgemm(CblasNoTrans, CblasConjTrans, alpha, param->do_a[i], param->do_a[i], beta, param->do_aad[i]);
                }
                if (param->do_g1[i] != 0.0)
                {
                        param->do_ada[i] = gsl_matrix_complex_alloc(ns, ns);
                        gsl_blas_zgemm(CblasConjTrans, CblasNoTrans, alpha, param->do_a[i], param->do_a[i], beta, param->do_ada[i]);
                }
        }
	
	/*
 	 * Add all operators to L:
 	 *
 	 */ 

	// L += -i H rho
	GSL_SET_COMPLEX(&z, 0.0, 1.0);
	steady_state_L_add_op(sop_L, param->H0, NULL, z, ns);

	// L +=  i rho H
	GSL_SET_COMPLEX(&z, 0.0, -1.0);
	steady_state_L_add_op(sop_L, NULL, param->H0, z, ns);

	// loop through all dissipation operators
	// drho_t += g(N+1)( a rho ad - 1/2ad a rho - 1/2rho ad a) + g N (ad rho a - 1/2a ad rho - 1/2rho a ad)
	for (i = 0; i < param->do_n; i++)
	{
		if (param->do_g1[i] != 0.0)
		{
			GSL_SET_COMPLEX(&z, -0.5 * param->do_g1[i], 0.0);
        		steady_state_L_add_op(sop_L, NULL, param->do_ada[i], z, ns);
        		steady_state_L_add_op(sop_L, param->do_ada[i], NULL, z, ns); // XXX: conjugate here...
			GSL_SET_COMPLEX(&z,  1.0 * param->do_g1[i], 0.0);
        		steady_state_L_add_op(sop_L, param->do_a[i], param->do_ad[i], z, ns);
		}
		if (param->do_g2[i] != 0.0)
		{
			GSL_SET_COMPLEX(&z, -0.5 * param->do_g2[i], 0.0);
        		steady_state_L_add_op(sop_L, NULL, param->do_aad[i], z, ns);
        		steady_state_L_add_op(sop_L, param->do_aad[i], NULL, z, ns); // XXX: conjugate here...
			GSL_SET_COMPLEX(&z,  1.0 * param->do_g2[i], 0.0);
        		steady_state_L_add_op(sop_L, param->do_ad[i], param->do_a[i], z, ns);
		}
	}

        //printf("L_imag =\n");
        //misc_matrix_complex_print_imag(sop_L, ns*ns);
        //printf("L_real =\n");
        //misc_matrix_complex_print_real(sop_L, ns*ns);

	/*
	 * Replace first row with density matrix normalisation condition:
	 *
	 */
	GSL_SET_COMPLEX(&z, 1.0, 0.0);
	for (i = 0; i < ns; i++)
	{
		gsl_matrix_complex_set(sop_L, 0, ss_idx_major(ns, i,i), z);
	}
	// and create rhs: 
	gsl_vector_complex_set_zero(ss_eq_rhs);
	gsl_vector_complex_set(ss_eq_rhs, 0, z);
	
	/*
 	 * Solve for rho_ss:
 	 *
 	 * L * rho_ss = [1, 0, 0, ...]'
 	 *
 	 */
        //printf("Lm_real =\n");
        //misc_matrix_complex_print_real(sop_L, ns*ns);
        //printf("Lm_imag =\n");
        //misc_matrix_complex_print_imag(sop_L, ns*ns);

       	gsl_linalg_complex_LU_decomp(sop_L, p, &s);
       	gsl_linalg_complex_LU_solve (sop_L, p, ss_eq_rhs, rho_ss_vec);

	//printf("rho_ss_vec = \n");
        //gsl_vector_complex_fprintf(stdout, rho_ss_vec, "%g");
	//printf("ss_eq_rhs = \n");
        //gsl_vector_complex_fprintf(stdout, ss_eq_rhs, "%g");


	/*
 	 * Copy rho_ss_vec to rho_ss:
 	 *
 	 */
	for (i = 0; i < ns*ns; i++)
	{
		z = gsl_vector_complex_get(rho_ss_vec, i);
		gsl_matrix_complex_set(rho_ss, ss_idx_minor_ms(ns, i), ss_idx_minor_ls(ns, i), z);
	}

        //printf("rho_ss_real =\n");
        //misc_matrix_complex_print_real(rho_ss, ns);

	printf("DEBUG: solving for rho_ss: finished\n");
		
	return 0;
}

