/*
 * Simulate the time evolution of a N-TLS system.
 *
 *
 * Robert Johansson <robert@riken.jp>
 *
 * $Id: run_2ho_diag.c,v 1.1.1.1 2008-09-03 09:36:41 rob Exp $
 */
#include "quantum_system.h"
#include "density_matrix.h"
#include "operators.h"
#include "hamiltonian.h"
#include "master_equation.h"
#include "gsl_ext.h"
#include "param.h"
#include "misc.h"

#include <math.h>
#include <stdio.h> 
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_eigen.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_cblas.h>
#include <gsl/gsl_complex.h>
#include <gsl/gsl_complex_math.h>

/* ---
 * Simulation parameters.
 * 
 */
double epsilon[] = { 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0};
double delta[]   = { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0};
double ho_w[]    = { 1.0, 0.01, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0};
double *lambda[10];

int
find_evec_from_qs_overlap(quantum_system *qs, int qsn, gsl_matrix_complex *H, gsl_matrix_complex *evec, gsl_vector *eval, int n, int m)
{
	double max_overlap = 0.0;
	int i, max_overlap_i = -1;
	quantum_system_state_vector_t qsv;
	gsl_complex z;
	
	gsl_vector_complex *v = gsl_vector_complex_alloc(qsn);
	
	qsv[0] = n; qsv[1] = m; 
	i = quantum_system_state_number(qs, qsv);
	gsl_vector_complex_set_zero(v);
	gsl_vector_complex_set(v, i, GSL_COMPLEX_ONE);

	// Find closest eigenvalue
	for (i = 0; i < qsn; i++)
	{
		gsl_vector_complex_view evec_i;
		evec_i = gsl_matrix_complex_column(evec, i);
		gsl_blas_zdotu(v, &evec_i.vector, &z);
		
		if (gsl_complex_abs(z) > max_overlap)
		{
			max_overlap = gsl_complex_abs(z);
			max_overlap_i = i;
		}
	}
	gsl_vector_complex_free(v);

	return max_overlap_i;
}

int
find_evec_from_qs(quantum_system *qs, int qsn, gsl_matrix_complex *H, gsl_matrix_complex *evec, gsl_vector *eval, int n, int m)
{
	double min_dev;
	int i, min_dev_i = -1;
	quantum_system_state_vector_t qsv;
	gsl_complex z;
	
	gsl_vector_complex *v = gsl_vector_complex_alloc(qsn);
	gsl_vector_complex *y = gsl_vector_complex_alloc(qsn);
	
	qsv[0] = n; qsv[1] = m; 
	i = quantum_system_state_number(qs, qsv);
	gsl_vector_complex_set_zero(v);
	gsl_vector_complex_set(v, i, GSL_COMPLEX_ONE);

	gsl_blas_zgemv(CblasNoTrans, GSL_COMPLEX_ONE, H, v, GSL_COMPLEX_ZERO, y);
	gsl_blas_zdotu(v, y, &z);

	// Find closest eigenvalue
	min_dev = 1e6;
	for (i = 0; i < qsn; i++)
	{
		double eval_i = gsl_vector_get(eval, i);
		if (min_dev > fabs(eval_i -  GSL_REAL(z)))
		{
			min_dev = fabs(eval_i -  GSL_REAL(z));
			min_dev_i = i;
		}
	}

	gsl_vector_complex_free(v);
	gsl_vector_complex_free(y);
	
//	printf("IDX DEBUG (E): |%d,%d> -> %d\n", n, m, min_dev_i);
	
	return min_dev_i;
}

/* ---
 * Program starts here.
 * 
 */
int
main(int argc, char **argv)
{
	quantum_system *qs;
	gsl_matrix_complex *H, *H_drive, *M;
	int qsn, i;
	double ho_g;

	solver_param param;

	ho_w[0] = strtod(argv[1], NULL) / (2*M_PI);
	ho_w[1] = strtod(argv[2], NULL) / (2*M_PI);
	ho_g = strtod(argv[3], NULL) / (2*M_PI);
	
	int n, m, j;
	gsl_vector *eval;
	gsl_vector_complex *y, *x;
	gsl_matrix_complex *evec;
	
	/*
	 * Setup solver parameters.
	 *
	 */
	param.epsilon = epsilon;
	param.delta   = delta;
	param.ho_w    = ho_w;
	for (i = 0; i < 8; i++)
	{
		param.epsilon[i] *= 2*M_PI;
		param.delta[i]   *= 2*M_PI;
		param.ho_w[i]    *= 2*M_PI;
		lambda[i] = (double *)calloc(10, sizeof(double));
	}
	lambda[0][1] = 0.25 * 2*M_PI;
	lambda[1][2] = 0.4 * 2*M_PI;
	lambda[2][3] = 0.3 * 2*M_PI;
	lambda[3][4] = 0.2 * 2*M_PI;
	lambda[4][5] = 0.1 * 2*M_PI;
	
	param.lambda  = lambda;

	param.ho_g = ho_g * 2*M_PI;
	
	/* 
	 * create a new quantum system object
	 */
	qs = quantum_system_new();
	quantum_system_add(qs, 15);
	quantum_system_add(qs, 15);
	qsn = quantum_system_nstates(qs);
	quantum_system_print(qs);


	/*
	 * Calculate hamiltonian
	 *
	 */
	H       = hamiltonian_2ho(qs, &param);
	H_drive = hamiltonian_2ho_drive(qs, &param);

	//printf("H_real =\n"); misc_matrix_complex_print_real(H, qsn);
	//printf("H_drive_real =\n"); misc_matrix_complex_print_real(H_drive, qsn);
	//printf("H_imag =\n"); misc_matrix_complex_print_imag(H, qsn);

	/*
	 * Diagonalize hamiltonian.
	 *
	 */
  	M = gsl_matrix_complex_alloc(qsn, qsn);
    	evec = gsl_matrix_complex_alloc(qsn, qsn);
      	eval = gsl_vector_alloc(qsn);
        gsl_eigen_hermv_workspace *w = gsl_eigen_hermv_alloc(qsn);

	gsl_matrix_complex_memcpy(M, H);
	gsl_eigen_hermv(M, eval, evec, w);
	gsl_eigen_hermv_sort(eval, evec, GSL_EIGEN_SORT_ABS_ASC);
	//test_eigen_complex_results(qsn, H, eval, evec);
	
	/*
	 * Check eigenvalues and eigenvectors
	 *
	 */
	printf("evals =\n");
	for (i = 0; i < qsn; i++)
	{
		double eval_i = gsl_vector_get(eval, i);
		printf("\t%f", eval_i);
	}
	printf("\n");

	//printf("evec_real = \n"); misc_matrix_complex_print(evec, qsn);
		
	x = gsl_vector_complex_alloc(qsn);
	y = gsl_vector_complex_alloc(qsn);

	for (i = 0; i < qsn; i++)
	{
		gsl_complex z;
		double eval_i, em_err = 0.0;
		gsl_vector_complex_const_view evec_i  = gsl_matrix_complex_const_column(evec, i);
		
		eval_i = gsl_vector_get(eval, i);
		gsl_vector_complex_memcpy(x, &evec_i.vector);

		gsl_blas_zgemv(CblasNoTrans, GSL_COMPLEX_ONE, H, x, GSL_COMPLEX_ZERO, y);

		printf("l = %f\n", eval_i);
		printf("Hv = ");
		for (j = 0; j < 8; j++)
		{
			z = gsl_vector_complex_get(y, j);
			printf("\t %f", GSL_REAL(z));
		}
		printf("\n");
		printf("lv = ");
		for (j = 0; j < 8; j++)
		{
			z = gsl_vector_complex_get(x, j);
			printf("\t %f", eval_i * GSL_REAL(z));
		}
		printf("\n");
		
		for (j = 0; j < qsn; j++)
		{
			gsl_complex z1, z2;
			double z_abs;
			
			z1 = gsl_vector_complex_get(y, j);
			z2 = gsl_vector_complex_get(x, j);
			z2 = gsl_complex_mul_real(z2, eval_i);
			z1 = gsl_complex_sub(z2, z1);	
			z_abs = gsl_complex_abs(z1);
			if (z_abs > em_err)
				em_err = z_abs;
		}
	
		
		printf("eigenvec/val %d: error = %g\n", i, em_err);

	}


	/*
	 * Calculate matrix elements
	 * 
	 */
	
	printf("<n,m|n,m>\n   ");
	//for (m = 0; m < qs->nstates[1]; m++) printf("\t\tm=%d\t", m);
	for (m = 0; m < 8; m++) printf("\t\tm=%d\t", m);
	printf("\n");
	
	//for (n = 0; n < qs->nstates[0]; n++)
	for (n = 0; n < 8; n++)
	{
		printf("n=%d", n);
		//for (m = 0; m < qs->nstates[1]; m++)
		for (m = 0; m < 8; m++)
		{
			gsl_complex z;
			gsl_vector_complex_view evec_i, evec_j;

			i = find_evec_from_qs(qs, qsn, H, evec, eval, n, m);
			j = find_evec_from_qs(qs, qsn, H, evec, eval, n, m);
			
			evec_i = gsl_matrix_complex_column(evec, i);
			evec_j = gsl_matrix_complex_column(evec, j);
			
			gsl_blas_zdotu(&evec_i.vector, &evec_j.vector, &z);
			
			printf("\t(%.5f, %.5f)", GSL_REAL(z), GSL_IMAG(z));
		}
		printf("\n");
	}


	printf("<n,m|H_drive|n+1,m-1> / (sqrt(m*(n+1)) 2 g / w_c)\n   ");
	//for (m = 0; m < qs->nstates[1]-1; m++) printf("\t\tm=%d\t", m);
	for (m = 0; m < 8; m++) printf("\t\tm=%d\t", m);
	printf("\n");
	//for (n = 0; n < qs->nstates[0]-1; n++)
	for (n = 0; n < 8; n++)
	{
		printf("n=%d", n);
		//for (m = 1; m < qs->nstates[1]; m++)
		for (m = 1; m < 8; m++)
		{
			gsl_complex z;
			gsl_vector_complex_view evec_i, evec_j;

			i = find_evec_from_qs(qs, qsn, H, evec, eval, n, m);
			j = find_evec_from_qs(qs, qsn, H, evec, eval, n+1, m-1);

			evec_i = gsl_matrix_complex_column(evec, i);
			evec_j = gsl_matrix_complex_column(evec, j);

			gsl_blas_zgemv(CblasNoTrans, GSL_COMPLEX_ONE, H_drive, &evec_j.vector, GSL_COMPLEX_ZERO, y);
			gsl_blas_zdotu(&evec_i.vector, y, &z);
			printf("\t(%.5f, %.5f)", GSL_REAL(z) / (sqrt((n+1)*m) * 2*ho_g / ho_w[1]), GSL_IMAG(z));
		}
		printf("\n");
	}

	exit(1);
	printf("<n,m|H_drive|n-1,m+2> / sqrt(n * (m+1) * (m+2)) / (g / w_c)^2 \n   "); // sqrt(n+1)/sqrt(m*(m-1))\n   ");
	for (m = 0; m < qs->nstates[1]-2; m++) printf("\t\tm=%d\t", m);
	printf("\n");
	for (n = 1; n < qs->nstates[0]; n++)
	{
		printf("n=%d", n);
		for (m = 0; m < qs->nstates[1]-2; m++)
		{
			gsl_complex z;
			gsl_vector_complex_view evec_i, evec_j;

			i = find_evec_from_qs(qs, qsn, H, evec, eval, n, m);
			j = find_evec_from_qs(qs, qsn, H, evec, eval, n-1, m+2);

			evec_i = gsl_matrix_complex_column(evec, i);
			evec_j = gsl_matrix_complex_column(evec, j);

			gsl_blas_zgemv(CblasNoTrans, GSL_COMPLEX_ONE, H_drive, &evec_j.vector, GSL_COMPLEX_ZERO, y);
			gsl_blas_zdotu(&evec_i.vector, y, &z);
			printf("\t(%.5f, %.5f)", GSL_REAL(z)/ sqrt(n * (m+1) * (m+2))  / (ho_g / ho_w[1])  / (ho_g / ho_w[1]), GSL_IMAG(z));
		}
		printf("\n");
	}

	printf("<n,m|H_drive|n-2,m+1> / sqrt(n*(n-1)*(m+1))\n   ");
	for (m = 0; m < qs->nstates[1]-1; m++) printf("\t\tm=%d\t", m);
	printf("\n");
	for (n = 2; n < qs->nstates[0]; n++)
	{
		printf("n=%d", n);
		for (m = 0; m < qs->nstates[1]-1; m++)
		{
			gsl_complex z;
			gsl_vector_complex_view evec_i, evec_j;

			i = find_evec_from_qs(qs, qsn, H, evec, eval, n, m);
			j = find_evec_from_qs(qs, qsn, H, evec, eval, n-2, m+1);

			evec_i = gsl_matrix_complex_column(evec, i);
			evec_j = gsl_matrix_complex_column(evec, j);

			gsl_blas_zgemv(CblasNoTrans, GSL_COMPLEX_ONE, H_drive, &evec_j.vector, GSL_COMPLEX_ZERO, y);
			gsl_blas_zdotu(&evec_i.vector, y, &z);
			printf("\t(%.5f, %.5f)", GSL_REAL(z)/ sqrt(n*(n-1)*(m+1)), GSL_IMAG(z));
		}
		printf("\n");
	}
	
	printf("<n,m|H_drive|n-1,m>/sqrt(n)\n   ");
	for (m = 0; m < qs->nstates[1]; m++) printf("\t\tm=%d\t", m);
	printf("\n");
	for (n = 1; n < qs->nstates[0]; n++)
	{
		printf("n=%d", n);
		for (m = 0; m < qs->nstates[1]; m++)
		{
			gsl_complex z;
			gsl_vector_complex_view evec_i, evec_j;

			i = find_evec_from_qs(qs, qsn, H, evec, eval, n, m);
			j = find_evec_from_qs(qs, qsn, H, evec, eval, n-1, m);

			evec_i = gsl_matrix_complex_column(evec, i);
			evec_j = gsl_matrix_complex_column(evec, j);

			gsl_blas_zgemv(CblasNoTrans, GSL_COMPLEX_ONE, H_drive, &evec_j.vector, GSL_COMPLEX_ZERO, y);
			gsl_blas_zdotu(&evec_i.vector, y, &z);
			printf("\t(%.5f, %.5f)", GSL_REAL(z)/sqrt(n), GSL_IMAG(z));
		}
		printf("\n");
	}
	
	printf("<n,m|H_drive|n-2,m+2>\n   ");
	for (m = 0; m < qs->nstates[1]-2; m++) printf("\t\tm=%d\t", m);
	printf("\n");
	for (n = 2; n < qs->nstates[0]; n++)
	{
		printf("n=%d", n);
		for (m = 0; m < qs->nstates[1]-2; m++)
		{
			gsl_complex z;
			gsl_vector_complex_view evec_i, evec_j;

			i = find_evec_from_qs(qs, qsn, H, evec, eval, n, m);
			j = find_evec_from_qs(qs, qsn, H, evec, eval, n-2, m+2);

			evec_i = gsl_matrix_complex_column(evec, i);
			evec_j = gsl_matrix_complex_column(evec, j);

			gsl_blas_zgemv(CblasNoTrans, GSL_COMPLEX_ONE, H_drive, &evec_j.vector, GSL_COMPLEX_ZERO, y);
			gsl_blas_zdotu(&evec_i.vector, y, &z);
			printf("\t(%.5f, %.5f)", GSL_REAL(z), GSL_IMAG(z));
		}
		printf("\n");
	}

	/*
	for (n = 0; n < qs->nstates[0]; n++)
	{
		for (m = 0; m < qs->nstates[1]; m++)
		{
			find_evec_from_qs_overlap(qs, qsn, H, evec, eval, n, m);
			find_evec_from_qs(qs, qsn, H, evec, eval, n, m);
		}
	}
	*/
	return 0;
}



