/*
 * Calculate the time evolution of the qubit + cavity.
 *
 * Emission spectrum calculation
 *
 * Robert Johansson <robert@riken.jp>
 *
 * $Id: run_ho.c,v 1.1.1.1 2008-09-03 09:36:41 rob Exp $
 */
#include "quantum_system.h"
#include "density_matrix.h"
#include "operators.h"
#include "hamiltonian.h"
#include "master_equation.h"
#include "gsl_ext.h"
#include "param.h"
#include "misc.h"

#include <math.h>
#include <stdio.h> 
#include <gsl/gsl_complex_math.h>
#include <gsl/gsl_complex.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_cblas.h>
#include <gsl/gsl_blas.h>

#define DATADIR "data/ho"

/* ---
 * Evaluate the correlation function <A(t+tau)A(t)> at tau
 *
 */
int
rho_corr_cb(gsl_matrix_complex *op_tau, double tau, quantum_system *qs, solver_param *sp)
{
        gsl_complex corr;
        char filename[1024], row[1024];

        corr = density_matrix_expectation_value(sp->A, op_tau);

        snprintf(filename, sizeof(filename), "%s/C%s.dat", DATADIR, sp->simsig);
        snprintf(row, sizeof(row), "%f\t%f\t%f", tau, GSL_REAL(corr), GSL_IMAG(corr));
        misc_file_row_add(filename, row);

        return 0;
}

/* ---
 * Store the final (steady-state) density matrix to file.
 */
int
rho_store_final_cb(gsl_matrix_complex *rho_t, double t, quantum_system *qs, solver_param *sp)
{
	char filename[128];
	
	snprintf(filename, sizeof(filename), "%s/rho_ss_%s.dat", DATADIR, sp->simsig);

	misc_matrix_write(filename, rho_t, rho_t->size1);
	
	return 0;
}
	
/* ---
 * Process the density matrix for time t (calc. expectations values, 
 * store to file, etc.) 
 */
int
rho_store_cb(gsl_matrix_complex *rho_t, double t, quantum_system *qs, solver_param *sp)
{
	int i, j, ri;
	gsl_complex op_expt;
	char filename[1024], row[16384];
	gsl_matrix_complex *dm_part;

	/*
	 * Store the density matrix at time t:
 	 */ 
	//ri = 0;	
	//snprintf(filename, sizeof(filename), "%s/rho_diag%s.dat", DATADIR, sp->simsig);
	//ri = snprintf(row, sizeof(row), "%f", t);
	//for (j = 0; j < rho_t->size1; j++)
	//{
	// 	ri += snprintf(&row[ri], sizeof(row)-ri, "\t%f", (double)GSL_REAL(gsl_matrix_complex_get(rho_t, j, j)));
	//	if (ri >= sizeof(row))
	//	{
	//		fprintf(stderr, "Error: Preparing row for %s overflowed.\n", sp->simsig);
	//		break;
	//	}
	//}
	//misc_file_row_add(filename, row);


	/*
	 * For each sub-system:
	 *
	 */
	for (i = 0; i < qs->nsubsys; i++)
        {
		ri = 0;	
		
		snprintf(filename, sizeof(filename), "%s/rho_ho_%d_diag%s.dat", DATADIR, i, sp->simsig);
		
                dm_part = density_matrix_traceout(qs, rho_t, i);

		ri = snprintf(row, sizeof(row), "%f", t);

		for (j = 0; j < qs->nstates[i]; j++)
		{
		 	ri += snprintf(&row[ri], sizeof(row)-ri, "\t%f", (double)GSL_REAL(gsl_matrix_complex_get(dm_part, j, j)));
		}
		misc_file_row_add(filename, row);
		
		/* -- expectation values -- */ 
		op_expt = density_matrix_expectation_value(rho_t, sp->N_op[i]);
		snprintf(filename, sizeof(filename), "%s/N_expt_%d%s.dat", DATADIR, i, sp->simsig);
		snprintf(row, sizeof(row), "%f\t%f\t%f\t%f", t, GSL_REAL(op_expt), GSL_IMAG(op_expt), sp->do_g1[0]);
		misc_file_row_add(filename, row);		
		
		op_expt = density_matrix_expectation_value(rho_t, sp->a);
		snprintf(filename, sizeof(filename), "%s/a_expt_%d%s.dat", DATADIR, i, sp->simsig);
		snprintf(row, sizeof(row), "%f\t%f\t%f", t, GSL_REAL(op_expt), GSL_IMAG(op_expt));
		misc_file_row_add(filename, row);		

		op_expt = density_matrix_expectation_value(rho_t, sp->ad);
		snprintf(filename, sizeof(filename), "%s/ad_expt_%d%s.dat", DATADIR, i, sp->simsig);
		snprintf(row, sizeof(row), "%f\t%f\t%f", t, GSL_REAL(op_expt), GSL_IMAG(op_expt));
		misc_file_row_add(filename, row);		

		
		//gsl_matrix_complex_free(dm_part);
        }
		
	/* -- expectation values -- */
	op_expt = density_matrix_expectation_value(rho_t, sp->A);
	snprintf(filename, sizeof(filename), "%s/A_expt%s.dat", DATADIR, sp->simsig);
	snprintf(row, sizeof(row), "%f\t%f\t%f", t, GSL_REAL(op_expt), GSL_IMAG(op_expt));
	misc_file_row_add(filename, row);		
		
	if (t >= sp->Tf-sp->dT)
	{
		printf("Storing final rho: rho_f\n");
	 	gsl_matrix_complex_memcpy(sp->rho_f, rho_t);
	}
	
	return 0;
}


/* ---
 * Simulation parameters.
 * 
 */
double ho_w[]    = { 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0};
double *lambda[10];


/* ---
 * Program starts here.
 * 
 */
int
main(int argc, char **argv)
{
	quantum_system *qs;
	gsl_matrix_complex *rho_q, *rho_c;
	gsl_matrix_complex *rho0, *rho_t;
	density_matrix_list_t dm_list;
	int qsn, i, noff, nsize;
	double Navg, fw, fA, phi, Gc, Gq, gqc, wT;

	solver_param param;

	if (argc != 6)
	{
		printf("usage: %s check code\n", argv[0]);
		exit(0);
	}
	
	nsize   = strtol(argv[1], NULL, 10);
	fw      = strtod(argv[2], NULL);
	fA      = strtod(argv[3], NULL);
	Gc	= strtod(argv[4], NULL);
	wT	= strtod(argv[5], NULL);

	/*
	 * Setup solver parameters.
	 *
	 */
	param.ho_w    = ho_w;
	for (i = 0; i < 10; i++)
	{
		param.ho_w[i]    *= 2*M_PI;
	}
	
	param.Ti = 0.0;
	param.dT = 0.2;
	param.Tf = 2000.0;
	param.n_offset = 0; 
	param.N = nsize; 

	printf("Using cavity with %d states, GND offset %d, and decay rate Gc = %f.\n", param.N, param.n_offset, Gc);
	
	param.f_A = fA * 2 * M_PI;
	param.f_w = fw * ho_w[0];

	param.H0_func = (hamiltonian_func_t)hamiltonian_ho;
	param.H1_func = (hamiltonian_func_t)hamiltonian_ho_drive;
	param.Ht_func = (hamiltonian_td_func_t)hamiltonian_t;
	
	/* 
	 * create a new quantum system object
	 */
	qs = quantum_system_new();
	quantum_system_add(qs, param.N);
	qsn = quantum_system_nstates(qs);
	quantum_system_print(qs);
	param.qs = qs;

	/*
	 * Set up dissipation
	 *
	 */
	param.wT = wT * 2 * M_PI;
	param.do_n = 0;

	/* --- cavity --- */
 	param.do_a[param.do_n] = operator_ho_lowering(qs, 0, param.n_offset); 
	if (param.wT > 0.0)
	{
		Navg = 1/(exp(param.ho_w[0]/param.wT) - 1);
	}
	else
	{
		Navg = 0;
	}
	param.do_g1[param.do_n] = Gc * (1 + Navg); // relaxation
	param.do_g2[param.do_n] = Gc * Navg;       // excitation
	param.do_n++;
	
	/* 
	 * Pre-calc of operators to calculate expectation values for.
	 */
	param.N_op[0] = operator_ho_N(qs, 0, param.n_offset);
	param.a  = operator_ho_lowering(qs, 0, param.n_offset);
	param.ad = operator_ho_raising(qs, 0, param.n_offset);
        param.A  = gsl_matrix_complex_alloc(qsn, qsn);
        gsl_matrix_complex_memcpy(param.A, param.a);
        gsl_matrix_complex_add(param.A, param.ad);
		
	snprintf(param.simsig, sizeof(param.simsig), "_Nc_%d_Noffset_%d_wc_%.3f_Gc_%.4f_fA_%.4f_fw_%.3f_wT_%.3f",
		qs->nstates[0], param.n_offset, param.ho_w[0]/(2*M_PI), Gc, param.f_A / (2 * M_PI), param.f_w / (2 * M_PI), param.wT / (2 * M_PI));

	/*
	 * Setup initial state
	 */
	rho_c = density_matrix_fock_state(10, qs->nstates[0]); 	// |0> state
        density_matrix_list_init(&dm_list);
        density_matrix_list_append(&dm_list, rho_c);
	rho0 = density_matrix_combine(qs, &dm_list);
        
	/* storage space for final rho */
	param.rho_f = gsl_matrix_complex_alloc(qsn, qsn);
	
	/*
	 * Evolve the quantum system until time t = T, where the steady state
         * is assumed to be reached.
	 * 
	 */
	printf("Start evolving system\n");
	if (dm_evolve_lme_t(qs, rho0, &param, rho_store_cb) != 0)
	//if (dm_evolve_he_t(qs, rho0, &param, rho_store_cb) != 0)
	//if (dm_evolve_unitary_const_full_steps(qs, rho0, &param, rho_store_cb) != 0)
	//if (dm_evolve_unitary_const_simple(qs, rho0, &param, rho_store_cb) != 0)
	//if (dm_evolve_unitary_const(qs, rho0, &param, rho_store_cb) != 0)
	//if (dm_evolve_unitary_t(qs, rho0, &param, rho_store_cb) != 0)
	{
		fprintf(stderr, "Evolution of the quantum system failed.\n");
		return -1;
	}

	/* 
         * At this point the system (rho_f) is supposed to be in a steady-state,
	 * such that C(t,tau) = C(tau).
	 *
	 */
	rho_store_final_cb(param.rho_f, param.Tf, qs, &param);

	/*
 	 * Read steady-state density matrix form file:
 	 */
	//param.rho_f = gsl_matrix_complex_alloc(qsn, qsn);
	//misc
	
	printf("Start calculating the correlation function\n");
        // here, assume that rho0 = rho(t), proceed to calculate Corr(t,tau)

        param.Ti = 0.00;
        param.Tf = 2000.00;       // tau
        param.dT = 0.05;

        rho_t = gsl_matrix_complex_alloc(qsn, qsn);
        gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, GSL_COMPLEX_ONE, param.A,  param.rho_f,  GSL_COMPLEX_ZERO, rho_t);
        //gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, GSL_COMPLEX_ONE, param.A,  rho0,  GSL_COMPLEX_ZERO, rho_t);

        /*
         * Evolve the B rho(t)
         *
         */
        if (dm_evolve_lme_t(qs, rho_t, &param, rho_corr_cb) != 0)
	//if (dm_evolve_unitary_const(qs, rho_t, &param, rho_corr_cb) != 0)
        //if (dm_evolve_unitary_const(qs, rho, &param, rho_store_cb) != 0)
        //if (dm_evolve_unitary_t(qs, rho, &param, rho_store_cb) != 0)
        {
                fprintf(stderr, "Evolution of the B rho failed.\n");
                return -1;
        }
	
	return 0;
}



