/*
 * Simulate the time evolution of a N-TLS system.
 *
 *
 * Robert Johansson <robert@riken.jp>
 *
 * $Id: run_ho_orig.c,v 1.1.1.1 2008-09-03 09:36:41 rob Exp $
 */
#include "quantum_system.h"
#include "density_matrix.h"
#include "operators.h"
#include "hamiltonian.h"
#include "master_equation.h"
#include "entanglement.h"
#include "gsl_ext.h"
#include "param.h"
#include "misc.h"

#include <math.h>
#include <stdio.h> 
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_cblas.h>
#include <gsl/gsl_blas.h>

/* ---
 * Process the density matrix for time t (calc. expectations values, 
 * store to file, etc.) 
 */
int
rho_store_cb(gsl_matrix_complex *rho_t, double t, quantum_system *qs, solver_param *sp)
{
	int i, j;
	gsl_complex N_expt; //, op_expt;
	char filename[128], row[1024];
	gsl_matrix_complex *dm_part;
	double wTeff[2], wS[2];
	
        for (i = 0; i < qs->nsubsys; i++)
        {
		int ri = 0;	
		
		snprintf(filename, sizeof(filename), "data.2ho_cool/rho_ho_%d_diag%s.dat", i, sp->simsig);
				
                dm_part = density_matrix_traceout(qs, rho_t, i);

		wS[i] = entanglement_neumann_entropy(dm_part);
		
		ri = snprintf(row, sizeof(row), "%f", t);

		for (j = 0; j < qs->nstates[i]; j++)
			 ri += snprintf(&row[ri], sizeof(row)-ri, "\t%f", GSL_REAL(gsl_matrix_complex_get(dm_part, j, j)));
	
		misc_file_row_add(filename, row);


		/* -- expectation values -- */
		N_expt = density_matrix_expectation_value(rho_t, sp->N_op[i]);
		snprintf(filename, sizeof(filename), "data.2ho_cool/N_expt_%d%s.dat", i, sp->simsig);
		snprintf(row, sizeof(row), "%f\t%f\t%f", t, GSL_REAL(N_expt), GSL_IMAG(N_expt));
		misc_file_row_add(filename, row);		
		
		sp->Nexpt[i] = GSL_REAL(N_expt);
/*
		op_expt = density_matrix_expectation_value(rho_t, sp->a);
		snprintf(filename, sizeof(filename), "data/a_expt_%d%s.dat", i, sp->simsig);
		snprintf(row, sizeof(row), "%f\t%f\t%f", t, GSL_REAL(op_expt), GSL_IMAG(op_expt));
		misc_file_row_add(filename, row);		

		op_expt = density_matrix_expectation_value(rho_t, sp->ad);
		snprintf(filename, sizeof(filename), "data/ad_expt_%d%s.dat", i, sp->simsig);
		snprintf(row, sizeof(row), "%f\t%f\t%f", t, GSL_REAL(op_expt), GSL_IMAG(op_expt));
		misc_file_row_add(filename, row);		

		op_expt = density_matrix_expectation_value(rho_t, sp->b);
		snprintf(filename, sizeof(filename), "data/b_expt_%d%s.dat", i, sp->simsig);
		snprintf(row, sizeof(row), "%f\t%f\t%f", t, GSL_REAL(op_expt), GSL_IMAG(op_expt));
		misc_file_row_add(filename, row);		

		op_expt = density_matrix_expectation_value(rho_t, sp->bd);
		snprintf(filename, sizeof(filename), "data/bd_expt_%d%s.dat", i, sp->simsig);
		snprintf(row, sizeof(row), "%f\t%f\t%f", t, GSL_REAL(op_expt), GSL_IMAG(op_expt));
		misc_file_row_add(filename, row);		

		op_expt = density_matrix_expectation_value(rho_t, sp->b2);
		snprintf(filename, sizeof(filename), "data/b2_expt_%d%s.dat", i, sp->simsig);
		snprintf(row, sizeof(row), "%f\t%f\t%f", t, GSL_REAL(op_expt), GSL_IMAG(op_expt));
		misc_file_row_add(filename, row);		

		op_expt = density_matrix_expectation_value(rho_t, sp->bd2);
		snprintf(filename, sizeof(filename), "data/bd2_expt_%d%s.dat", i, sp->simsig);
		snprintf(row, sizeof(row), "%f\t%f\t%f", t, GSL_REAL(op_expt), GSL_IMAG(op_expt));
		misc_file_row_add(filename, row);		
*/
		
		gsl_matrix_complex_free(dm_part);
        }
	
	/* --- store effective temperatures calculated above --- */
	wTeff[0] = sp->ho_w[0] / log(1/sp->Nexpt[0] + 1);
	wTeff[1] = sp->ho_w[1] / log(1/sp->Nexpt[1] + 1);
	snprintf(filename, sizeof(filename), "data.2ho_cool/wT_eff%s.dat", sp->simsig);
	snprintf(row, sizeof(row), "%f\t%f\t%f\t%f\t%f", t, wTeff[0], wTeff[1], wS[0], wS[1]);
	misc_file_row_add(filename, row);		

		
	return 0;
}


/* ---
 * Simulation parameters.
 * 
 */
double epsilon[] = { 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0};
double delta[]   = { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0};
double ho_w[]    = { 1.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0};
double *lambda[10];


/* ---
 * Program starts here.
 * 
 */
int
main(int argc, char **argv)
{
	quantum_system *qs;
	gsl_matrix_complex *rho1, *rho2;
	gsl_matrix_complex *rho0;
	density_matrix_list_t dm_list;
	int qsn, i;
	double Navg, wTeff[2], g0, fw, fA;

	solver_param param;

	if (argc != 5)
	{
		printf("usage: %s w1 fA g0\n", argv[0]);
	}
	
	fA      = strtod(argv[1], NULL);
	fw      = strtod(argv[2], NULL);
	g0      = strtod(argv[3], NULL);

	printf("Input arguments: ho_w[0] = %f, fA = %f, fw = %f\n", ho_w[0], fA, fw);
	
	/*
	 * Setup solver parameters.
	 *
	 */
	param.epsilon = epsilon;
	param.delta   = delta;
	param.ho_w    = ho_w;
	for (i = 0; i < 10; i++)
	{
		param.ho_w[i]    *= 2*M_PI;
	}
	
	param.Ti =  0.0;
	param.Tf =  100.0; 
	param.dT =  0.1;

	param.N = 2;
	param.f_A = fA * 2 * M_PI;
	param.f_w = fw * ho_w[0]; // XXX

	//param.H0_func = (hamiltonian_func_t)hamiltonian_nTLS;
	param.H0_func = (hamiltonian_func_t)hamiltonian_ho;
	param.H1_func = (hamiltonian_func_t)hamiltonian_ho_drive;
	
	/* 
	 * create a new quantum system object
	 */
	qs = quantum_system_new();
	//for (i = 0; i < param.N; i++)
	//	quantum_system_add(qs, 2);
	quantum_system_add(qs, 20);
	qsn = quantum_system_nstates(qs);
	quantum_system_print(qs);
	param.qs = qs;


	/*
	 * Set up dissipation
	 *
	 */
	param.wT = 0.5 * 2 * M_PI;
	param.do_n = 0;
	param.do_a[param.do_n] = operator_ho_lowering(qs, 0); // ho 0
	Navg = 1/(exp(param.ho_w[0]/param.wT) - 1);
	param.do_g1[param.do_n] = g0 * (1 + Navg); // relaxation
	param.do_g2[param.do_n] = g0 * Navg;       // excitation
	param.do_n++;

 	//param.do_a[param.do_n] = operator_ho_lowering(qs, 1); // ho 1
	//Navg = 1/(exp(param.ho_w[1]/param.wT) - 1);
	//param.do_g1[param.do_n] = 0.1 * (1 + Navg); // relaxation
	//param.do_g2[param.do_n] = 0.1 * Navg; // excitation
	//param.do_n++;

	/* 
	 * Pre-calc of operators to calculate expectation values for.
	 */
	param.N_op[0] = operator_ho_N(qs, 0);
	//param.N_op[1] = operator_ho_N(qs, 1);
	//param.a  = operator_ho_lowering(qs, 0);
	//param.ad = operator_ho_raising(qs, 0);
	//param.b  = operator_ho_lowering(qs, 1);
	//param.bd = operator_ho_raising(qs, 1);
	//param.b2  = gsl_matrix_complex_alloc(qsn, qsn);
	//param.bd2  = gsl_matrix_complex_alloc(qsn, qsn);
	//{
	//	gsl_complex alpha, beta;
	//	GSL_SET_COMPLEX(&alpha, 1.0, 0.0);
	//	GSL_SET_COMPLEX(&beta, 0.0, 0.0);
	//	gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, param.b,  param.b,  beta, param.b2);
	//	gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, param.bd, param.bd, beta, param.bd2);
	//}

	snprintf(param.simsig, sizeof(param.simsig), "_N_%d_w0_%.2f_wT_%.2f_fA_%.2f_fw_%.2f_g0_%.2f", 
		 qs->nstates[0], param.ho_w[0]/(2*M_PI), param.wT/(2*M_PI), param.f_A/(2*M_PI), param.f_w/(2*M_PI));
	
	/*
	 * Setup initial state
	 */
	//rho1 = density_matrix_pure_TLS(0.8, 0.2); 		// |> state
	//rho2 = density_matrix_pure_TLS(0.3, 0.7); 		// |> state
	rho1 = density_matrix_fock_state(0, qs->nstates[0]); 	// |0> state
	//rho2 = density_matrix_fock_state(0, qs->nstates[1]); 	// |0> state
	
        density_matrix_list_init(&dm_list);
        density_matrix_list_append(&dm_list, rho1);
        //for (i = 1; i < param.N; i++)
	//density_matrix_list_append(&dm_list, rho2);

	rho0 = density_matrix_combine(qs, &dm_list);

	/*
	 * Evolve the quantum system.
	 * 
	 */
	if (dm_evolve_lme_t(qs, rho0, &param, rho_store_cb) != 0)
	//if (dm_evolve_unitary_const(qs, rho0, &param, rho_store_cb) != 0)
	//if (dm_evolve_unitary_t(qs, rho0, &param, rho_store_cb) != 0)
	{
		fprintf(stderr, "Evolution of the quantum system failed.\n");
		return -1;
	}

	/*
	 * Calculate final effective temperatures for HOs.
	 *
	 * /
	{
		char filename[512], row[512];
		FILE *f;

		snprintf(filename, sizeof(filename), "data.2ho_cool/Teff%s.dat", param.simsig);
		f = fopen(filename, "w");
		
		wTeff[0] = param.ho_w[0] / log(1/param.Nexpt[0] + 1);
		wTeff[1] = param.ho_w[1] / log(1/param.Nexpt[1] + 1);

		fprintf(f, "wT = %f\n", param.wT);
		fprintf(f, "<N0(tf)> = %f\n", param.Nexpt[0]);
		fprintf(f, "<N1(tf)> = %f\n", param.Nexpt[1]);

		fprintf(f, "<N0_eq> = %f\n", 1/(exp(param.ho_w[0]/param.wT) - 1));
		fprintf(f, "<N1_eq> = %f\n", 1/(exp(param.ho_w[1]/param.wT) - 1));
	
		fprintf(f, "wTeff[0] = %f\n", wTeff[0]);
		fprintf(f, "wTeff[1] = %f\n", wTeff[1]);

		fprintf(f, "w[0] = %f\n", param.ho_w[0]);
		fprintf(f, "w[1] = %f\n", param.ho_w[1]);

		fprintf(f, "w[1] / w[0]\t= %f\n", param.ho_w[1] / param.ho_w[0]);
		fprintf(f, "wTeff[1] / wTeff[0]\t= %f\n", wTeff[1] / wTeff[0]);

		snprintf(filename, sizeof(filename), "data.2ho_cool/Teff.dat");
		snprintf(row, sizeof(row), "%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f", 
			 param.ho_w[1]/(2*M_PI), fA, param.f_w / (2*M_PI), g0, wTeff[0], wTeff[1], 
			 param.Nexpt[0], param.Nexpt[1], param.ho_w[1]/param.ho_w[0], wTeff[1]/wTeff[0],
			 param.ho_w[1]/param.ho_w[0] - wTeff[1]/wTeff[0]);
		misc_file_row_add(filename, row);	
	}
	*/
	
	return 0;
}



