/*
 * Simulate the time evolution of a system of N two-level systems (TLS).
 *
 * Robert Johansson <robert@riken.jp>
 *
 * $Id: run_nTLS.c,v 1.1.1.1 2008-09-03 09:36:41 rob Exp $
 */
#include "quantum_system.h"
#include "density_matrix.h"
#include "operators.h"
#include "hamiltonian.h"
#include "master_equation.h"
#include "gsl_ext.h"
#include "param.h"
#include "misc.h"

#include <math.h>
#include <stdio.h> 
#include <gsl/gsl_matrix.h>

/** 
 * Process the density matrix for time t (calc. expectations values, 
 * store to file, etc.). This is a callback function that the solver
 * call in each time-step.
 */
int
rho_store_cb(gsl_matrix_complex *rho_t, double t, quantum_system *qs, solver_param *sp)
{
	double c;
	int i;
	char filename[128], row[512];
	gsl_matrix_complex *dm_part;
	
        for (i = 0; i < qs->nsubsys; i++)
        {
		snprintf(filename, sizeof(filename), "data.nTLS/tls_%d_dynamics.dat", i);
				
                dm_part = density_matrix_traceout(qs, rho_t, i);

		snprintf(row, sizeof(row), "%f\t%f\t%f\t%f\t%f", t, 
		         GSL_REAL(gsl_matrix_complex_get(dm_part, 0, 0)),
		         GSL_REAL(gsl_matrix_complex_get(dm_part, 0, 1)),
		         GSL_REAL(gsl_matrix_complex_get(dm_part, 1, 0)),
		         GSL_REAL(gsl_matrix_complex_get(dm_part, 1, 1)));
		
		misc_file_row_add(filename, row);
                gsl_matrix_complex_free(dm_part);
        }

	/* --- concurrence -- */
	snprintf(filename, sizeof(filename), "data.nTLS/concurrence.dat");
	c = density_matrix_concurrence(qs, rho_t);
	snprintf(row, sizeof(row), "%f\t%f", t, c); 
	misc_file_row_add(filename, row);
	
	return 0;
}


/* ---
 * Two-level system parameters: 
 *
 * H = sum_i(-0.5 espilon[i] sigma_z(i) - 0.5 delta[i] sigma_x(i)) + sum_ij(- 0.5 lambda[i][j] sigma_x[i] sigma_x[j]) 
 * 
 */
double epsilon[] = { 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0};
double delta[]   = { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0};
double *lambda[10];


/* ---
 * Program starts here.
 * 
 */
int
main(int argc, char **argv)
{
	quantum_system *qs;
	gsl_matrix_complex *rho1, *rho2;
	gsl_matrix_complex *rho0;
	density_matrix_list_t dm_list;
	int qsn, i;

	solver_param param;

	/*
	 * Setup solver parameters.
	 *
	 */

	// TLS hamiltonian parameters
	param.epsilon = epsilon;
	param.delta   = delta;
	for (i = 0; i < 10; i++)
	{
		param.epsilon[i] *= 2*M_PI;
		param.delta[i]   *= 2*M_PI;
		lambda[i] = (double *)calloc(10, sizeof(double));
	}
	lambda[0][1] = 0.25 * 2*M_PI;
	lambda[1][2] = 0.4 * 2*M_PI;
	lambda[2][3] = 0.3 * 2*M_PI;
	lambda[3][4] = 0.2 * 2*M_PI;
	lambda[4][5] = 0.1 * 2*M_PI;
	param.lambda  = lambda;
	
	
	// time range
	param.Ti =  0.0;
	param.Tf =  10.0;
	param.dT =  0.01;

	param.N = 6;

	// driving field (set to zero to disable)
	param.f_A = 0.0 * 2 * M_PI;
	param.f_w = 0.0 * 2 * M_PI;


	// function pointer to the functions generating the hamiltonian matrices
	param.H0_func = (hamiltonian_func_t)hamiltonian_nTLS;
	param.H1_func = (hamiltonian_func_t)hamiltonian_drive;
	param.Ht_func = (hamiltonian_td_func_t)hamiltonian_t;

	/* 
	 * create a new quantum system object
	 */
	qs = quantum_system_new();

	for (i = 0; i < param.N; i++)
		quantum_system_add(qs, 2);
	qsn = quantum_system_nstates(qs);
	quantum_system_print(qs);


	/*
	 * Set up dissipation
	 *
	 */
	param.do_n = 0;
	param.do_a[param.do_n] = operator_ho_lowering(qs, 0); // qubit 0
	param.do_g1[param.do_n] = 0.50; // relaxation
	param.do_g2[param.do_n] = 0.05; // excitation
	param.do_n++;
	param.do_a[param.do_n] = operator_ho_lowering(qs, 1); // qubit 1
	param.do_g1[param.do_n] = 0.25; // relaxation
	param.do_g2[param.do_n] = 0.01; // excitation
	param.do_n++;
	param.do_a[param.do_n] = operator_sigma_z(qs, 0); // qubit 0
	param.do_g1[param.do_n] = 0.25; // dephasing
	param.do_n++;


	/*
	 * debug
	 *
	 */
	printf("sigma_x(0) = \n"); misc_matrix_complex_print(operator_sigma_x(qs,0), 4);
	printf("sigma_x(0) = \n"); misc_matrix_complex_print(operator_sigma_x(qs,1), 4);
	printf("sigma_y(0) = \n"); misc_matrix_complex_print(operator_sigma_y(qs,0), 4);
	printf("sigma_y(0) = \n"); misc_matrix_complex_print(operator_sigma_y(qs,1), 4);

	
	/*
	 * Setup initial state
	 */
	rho1 = density_matrix_pure_TLS(0.8); // 0.8|0>+0.2|1>
	rho2 = density_matrix_pure_TLS(0.3); // 0.3|0>+0.7|1> 
	
        density_matrix_list_init(&dm_list);
        density_matrix_list_append(&dm_list, rho1);
        for (i = 1; i < param.N; i++)
		density_matrix_list_append(&dm_list, rho2);

	rho0 = density_matrix_combine(qs, &dm_list);
	
	//printf("rho0 =\n");
	//misc_matrix_complex_print(rho0, qsn);


	/*
	 * Evolve the quantum system.
	 * 
	 */
	//if (dm_evolve_lme_t(qs, rho0, &param, rho_store_cb) != 0)
	if (dm_evolve_unitary_const(qs, rho0, &param, rho_store_cb) != 0)
	//if (dm_evolve_unitary_t(qs, rho0, &param, rho_store_cb) != 0)
	{
		fprintf(stderr, "Evolution of the quantum system failed.\n");
		return -1;
	}

	return 0;
}



