/*
 * Calculate the time evolution of the qubit + cavity + transmission
 * line system.
 *
 * Transmission spectrum calculation
 *
 * Robert Johansson <robert@riken.jp>
 *
 * $Id: run_qho_tms.c,v 1.1.1.1 2008-09-03 09:36:41 rob Exp $
 */
#include "quantum_system.h"
#include "density_matrix.h"
#include "operators.h"
#include "hamiltonian.h"
#include "master_equation.h"
#include "gsl_ext.h"
#include "param.h"
#include "misc.h"

#include <math.h>
#include <stdio.h> 
#include <gsl/gsl_complex_math.h>
#include <gsl/gsl_complex.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_cblas.h>
#include <gsl/gsl_blas.h>


/* ---
 * Evaluate the correlation function <A(t+tau)A(t)> at tau
 *
 */
int
rho_corr_cb(gsl_matrix_complex *op_tau, double tau, quantum_system *qs, solver_param *sp)
{
        int qsn;
        gsl_complex corr1, corr2;
        gsl_matrix_complex *c;
        char filename[128], row[1024];

        qsn = quantum_system_nstates(qs);
        c = gsl_matrix_complex_alloc(qsn, qsn);

        gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, GSL_COMPLEX_ONE, sp->A,  op_tau,  GSL_COMPLEX_ZERO, c);

        corr1 = operator_trace(c);

        gsl_matrix_complex_free(c);

        corr2 = density_matrix_expectation_value(sp->A, op_tau);

        snprintf(filename, sizeof(filename), "data.qho_tms/C%s.dat", sp->simsig);
        snprintf(row, sizeof(row), "%f\t%f\t%f\t%f\t%f", tau, GSL_REAL(corr1), GSL_IMAG(corr1), GSL_REAL(corr2), GSL_IMAG(corr2));
        misc_file_row_add(filename, row);

        return 0;
}

/* ---
 * Process the density matrix for time t (calc. expectations values, 
 * store to file, etc.) 
 */
int
rho_store_final_cb(gsl_matrix_complex *rho_t, double t, quantum_system *qs, solver_param *sp)
{
	gsl_complex N_expt; //, op_expt;
	char filename[128], row[1024];
	
	snprintf(filename, sizeof(filename), "data.qho_tms/rho_f_expt_%d.dat", sp->N);
	
	/* -- expectation values -- */
	N_expt = density_matrix_expectation_value(rho_t, sp->N_op[1]);
	snprintf(row, sizeof(row), "%f\t%f\t%f\t%f", sp->f_w/(2*M_PI), sp->f_A/(2*M_PI), GSL_REAL(N_expt), GSL_IMAG(N_expt));
	misc_file_row_add(filename, row);		

	return 0;
}
	
/* ---
 * Process the density matrix for time t (calc. expectations values, 
 * store to file, etc.) 
 */
int
rho_store_cb(gsl_matrix_complex *rho_t, double t, quantum_system *qs, solver_param *sp)
{
	int i, j, ri;
	gsl_complex N_expt; //, op_expt;
	char filename[128], row[1024];
	gsl_matrix_complex *dm_part; //, *op;

	ri = 0;	
	snprintf(filename, sizeof(filename), "data.qho_tms/rho_diag%s.dat", sp->simsig);
	ri = snprintf(row, sizeof(row), "%f", t);
	for (j = 0; j < rho_t->size1; j++)
	{
	 	ri += snprintf(&row[ri], sizeof(row)-ri, "\t%f", (double)GSL_REAL(gsl_matrix_complex_get(rho_t, j, j)));
	}
	misc_file_row_add(filename, row);
	
	for (i = 0; i < qs->nsubsys; i++)
        {
		ri = 0;	
		
		snprintf(filename, sizeof(filename), "data.qho_tms/rho_ho_%d_diag%s.dat", i, sp->simsig);
		
                dm_part = density_matrix_traceout(qs, rho_t, i);

		ri = snprintf(row, sizeof(row), "%f", t);

		for (j = 0; j < qs->nstates[i]; j++)
		{
		 	ri += snprintf(&row[ri], sizeof(row)-ri, "\t%f", (double)GSL_REAL(gsl_matrix_complex_get(dm_part, j, j)));
		}
		misc_file_row_add(filename, row);
		
		/* -- expectation values -- */ 
		N_expt = density_matrix_expectation_value(rho_t, sp->N_op[i]);
		snprintf(filename, sizeof(filename), "data.qho_tms/N_expt_%d%s.dat", i, sp->simsig);
		snprintf(row, sizeof(row), "%f\t%f\t%f", t, GSL_REAL(N_expt), GSL_IMAG(N_expt));
		misc_file_row_add(filename, row);		
		
		/* -- trace check values -- * /
		op = gsl_ext_matrix_convert_and_free(operator_unit(qs, i));
		N_expt = density_matrix_expectation_value(dm_part, op);
		snprintf(filename, sizeof(filename), "data.qho_tms/Trace_%d%s.dat", i, sp->simsig);
		snprintf(row, sizeof(row), "%f\t%f\t%f", t, GSL_REAL(N_expt), GSL_IMAG(N_expt));
		misc_file_row_add(filename, row);		
		gsl_matrix_complex_free(op);
		*/	
		
/*
		op_expt = density_matrix_expectation_value(rho_t, sp->a);
		snprintf(filename, sizeof(filename), "data.qho_corr/a_expt_%d%s.dat", i, sp->simsig);
		snprintf(row, sizeof(row), "%f\t%f\t%f", t, GSL_REAL(op_expt), GSL_IMAG(op_expt));
		misc_file_row_add(filename, row);		

		op_expt = density_matrix_expectation_value(rho_t, sp->ad);
		snprintf(filename, sizeof(filename), "data.qho_corr/ad_expt_%d%s.dat", i, sp->simsig);
		snprintf(row, sizeof(row), "%f\t%f\t%f", t, GSL_REAL(op_expt), GSL_IMAG(op_expt));
		misc_file_row_add(filename, row);		

		op_expt = density_matrix_expectation_value(rho_t, sp->b);
		snprintf(filename, sizeof(filename), "data.qho_corr/b_expt_%d%s.dat", i, sp->simsig);
		snprintf(row, sizeof(row), "%f\t%f\t%f", t, GSL_REAL(op_expt), GSL_IMAG(op_expt));
		misc_file_row_add(filename, row);		

		op_expt = density_matrix_expectation_value(rho_t, sp->bd);
		snprintf(filename, sizeof(filename), "data.qho_corr/bd_expt_%d%s.dat", i, sp->simsig);
		snprintf(row, sizeof(row), "%f\t%f\t%f", t, GSL_REAL(op_expt), GSL_IMAG(op_expt));
		misc_file_row_add(filename, row);		

		op_expt = density_matrix_expectation_value(rho_t, sp->b2);
		snprintf(filename, sizeof(filename), "data.qho_corr/b2_expt_%d%s.dat", i, sp->simsig);
		snprintf(row, sizeof(row), "%f\t%f\t%f", t, GSL_REAL(op_expt), GSL_IMAG(op_expt));
		misc_file_row_add(filename, row);		

		op_expt = density_matrix_expectation_value(rho_t, sp->bd2);
		snprintf(filename, sizeof(filename), "data.qho_corr/bd2_expt_%d%s.dat", i, sp->simsig);
		snprintf(row, sizeof(row), "%f\t%f\t%f", t, GSL_REAL(op_expt), GSL_IMAG(op_expt));
		misc_file_row_add(filename, row);		
*/
		
		//gsl_matrix_complex_free(dm_part);
        }
		
	/* -- expectation values -- * /
	N_expt = density_matrix_expectation_value(rho_t, sp->A);
	snprintf(filename, sizeof(filename), "data.qho_tms/E_expt%s.dat", sp->simsig);
	snprintf(row, sizeof(row), "%f\t%f\t%f", t, GSL_REAL(N_expt), GSL_IMAG(N_expt));
	misc_file_row_add(filename, row);		
	*/
		
	/* -- trace check values -- * /
	op = gsl_ext_matrix_convert_and_free(operator_unit(qs, -1));
	N_expt = density_matrix_expectation_value(rho_t, op);
	snprintf(filename, sizeof(filename), "data.qho_tms/Trace_full%s.dat", sp->simsig);
	snprintf(row, sizeof(row), "%f\t%f\t%f", t, GSL_REAL(N_expt), GSL_IMAG(N_expt));
	misc_file_row_add(filename, row);		
	gsl_matrix_complex_free(op);
	*/
	
	if (t >= sp->Tf-sp->dT)
	{
		printf("Storing final rho: rho_f\n");
	 	gsl_matrix_complex_memcpy(sp->rho_f, rho_t);
	}
	
	return 0;
}


/* ---
 * Simulation parameters.
 * 
 */
double epsilon[] = { 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0};
double delta[]   = { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0};
double ho_w[]    = {9.899, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0};
double *lambda[10];


/* ---
 * Program starts here.
 * 
 */
int
main(int argc, char **argv)
{
	quantum_system *qs;
	gsl_matrix_complex *rho_q, *rho_c;
	gsl_matrix_complex *rho0, *rho_t;
	density_matrix_list_t dm_list;
	int qsn, i;
	double Navg, fw, fA, phi, Gc, Gq, gqc;

	solver_param param;

	if (argc != 6)
	{
		printf("usage: %s check code\n", argv[0]);
		exit(0);
	}
	
	phi     = strtod(argv[1], NULL);
	fw      = strtod(argv[2], NULL);
	fA      = strtod(argv[3], NULL);
	gqc	= strtod(argv[4], NULL);
	Gq	= strtod(argv[5], NULL);

	printf("Input arguments: phi = %f\n", phi);
	
	/*
	 * Setup solver parameters.
	 *
	 */
	
	// the qubit
	epsilon[0] = 0;
	delta[0] = 9.899; //13.7 * cos(fabs(M_PI * phi));
	//epsilon[0] = 12.899;
	//delta[0] = 0;

	Gc = 0.5;	// cavity decay rate
	
	printf("Qubit parameters: eps = %f, delta = %f\n", epsilon[0], delta[0]);
	
	param.epsilon = epsilon;
	param.delta   = delta;
	param.ho_w    = ho_w;
	for (i = 0; i < 10; i++)
	{
		param.epsilon[i] *= 2*M_PI;
		param.delta[i]   *= 2*M_PI;
		param.ho_w[i]    *= 2*M_PI;
		lambda[i] = (double *)calloc(10, sizeof(double));
	}
	lambda[0][1] = 0.0 * 2*M_PI;
	
	param.lambda  = lambda;

	param.g_qc = gqc * 2*M_PI; // 0.01
	param.g_cr = 0.0 * 2*M_PI; // 0.01
	
	param.Ti =  0.0;
	param.Tf =  50.0;
	param.dT =  0.002;

	//fA = 0.00;
	param.N = 20;
	//fA = 0.25;
	//fw = 1.0;
	param.f_A = fA * 2 * M_PI;
	param.f_w = fw * ho_w[0];

	//param.H0_func = (hamiltonian_func_t)hamiltonian_nTLS;
	param.H0_func = (hamiltonian_func_t)hamiltonian_qho;
	param.H1_func = (hamiltonian_func_t)hamiltonian_qho_drive;
	param.Ht_func = (hamiltonian_td_func_t)hamiltonian_t;
	
	/* 
	 * create a new quantum system object
	 */
	qs = quantum_system_new();
	quantum_system_add(qs, 2);
	quantum_system_add(qs, param.N);
	qsn = quantum_system_nstates(qs);
	quantum_system_print(qs);
	param.qs = qs;


	/*
	 * Set up dissipation
	 *
	 */
	param.wT = 0.0 * 2 * M_PI;
	param.do_n = 0;
	Navg = 0;

	/* --- qubit --- */
	//param.do_a[param.do_n] = operator_ho_lowering(qs, 0); 
	param.do_a[param.do_n] = operator_ho_raising(qs, 0); 
	Navg = 1/(exp(sqrt(param.epsilon[0]*param.epsilon[0] + param.delta[0]*param.delta[0])/param.wT) - 1);
	param.do_g1[param.do_n] = Gq * (1 + Navg); // relaxation
	param.do_g2[param.do_n] = Gq * Navg;       // excitation
	param.do_n++;
		
	/* --- cavity --- */
 	param.do_a[param.do_n] = operator_ho_lowering(qs, 1); 
	Navg = 1/(exp(param.ho_w[0]/param.wT) - 1);
	param.do_g1[param.do_n] = Gc * (1 + Navg); // relaxation
	param.do_g2[param.do_n] = Gc * Navg;       // excitation
	param.do_n++;
	
	/* --- transmission line --- * /
 	param.do_a[param.do_n] = operator_ho_lowering(qs, 2); 
	Navg = 1/(exp(param.ho_w[1]/param.wT) - 1);
	param.do_g1[param.do_n] = 0.05 * (1 + Navg); // relaxation
	param.do_g2[param.do_n] = 0.05 * Navg;       // excitation
	param.do_n++;
	*/

	/* 
	 * Pre-calc of operators to calculate expectation values for.
	 */
	//param.N_op[0] = operator_ho_N(qs, 0);
	//printf("N0 =\n");
	//misc_matrix_complex_print(param.N_op[0], param.N_op[0]->size1);
	//param.N_op[1] = operator_ho_N(qs, 1);
	//printf("N1 =\n");
	//misc_matrix_complex_print(param.N_op[1], param.N_op[1]->size1);
//	param.N_op[2] = gsl_ext_matrix_convert_and_free(operator_unit(qs, -1)); //operator_ho_N(qs, 2);
/*
	param.a  = operator_ho_lowering(qs, 0);
	param.ad = operator_ho_raising(qs, 0);
	param.b  = operator_ho_lowering(qs, 1);
	param.bd = operator_ho_raising(qs, 1);
	param.b2  = gsl_matrix_complex_alloc(qsn, qsn);
	param.bd2  = gsl_matrix_complex_alloc(qsn, qsn);
	{
		gsl_complex alpha, beta;
		GSL_SET_COMPLEX(&alpha, 1.0, 0.0);
		GSL_SET_COMPLEX(&beta, 0.0, 0.0);
		gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, param.b,  param.b,  beta, param.b2);
		gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, param.bd, param.bd, beta, param.bd2);
	}
*/
		
	snprintf(param.simsig, sizeof(param.simsig), "_NEC_Nq_%d_Nc_%d_Nr_%d_wq_%.2f_wc_%.2f_wr_%.2f_gqc_%.2f_Gc_%.2f_qdiss",
		qs->nstates[0], qs->nstates[1], qs->nstates[2],
		sqrt(param.epsilon[0]*param.epsilon[0] + param.delta[0]*param.delta[0])/(2*M_PI), param.ho_w[0]/(2*M_PI), param.ho_w[1]/(2*M_PI),
		param.g_qc/(2*M_PI), 1000*Gc);

	/*
	 * Setup initial state
	 */
	//rho1 = density_matrix_pure_TLS(0.8, 0.2); 		// |> state
	rho_q = density_matrix_pure_TLS(0.1); 		// |> state
	//rho_c = density_matrix_fock_state(0, qs->nstates[1]); 	// |0> state
	rho_c = density_matrix_coherent_state(2, 0, qs->nstates[1]); 	// |0> state
	//rho_r = density_matrix_fock_state(0, qs->nstates[2]); 	// |0> state
	
        density_matrix_list_init(&dm_list);
        density_matrix_list_append(&dm_list, rho_q);
        density_matrix_list_append(&dm_list, rho_c);
	rho0 = density_matrix_combine(qs, &dm_list);
	//printf("rho0 =\n");
	//misc_matrix_complex_print(rho0, rho0->size1);
        
	/* storage space for final rho */
	param.rho_f = gsl_matrix_complex_alloc(qsn, qsn);

	param.N_op[0] = operator_ho_N(qs, 0);
	//printf("N0 =\n");
	//misc_matrix_complex_print(param.N_op[0], param.N_op[0]->size1);
	param.N_op[1] = operator_ho_N(qs, 1);
	//printf("N1 =\n");
	//misc_matrix_complex_print(param.N_op[1], param.N_op[1]->size1);
        
	param.a  = operator_ho_lowering(qs, 1);
        param.ad = operator_ho_raising(qs, 1);
        param.A = gsl_matrix_complex_alloc(qsn, qsn);
        gsl_matrix_complex_memcpy(param.A, param.a);
        gsl_matrix_complex_add(param.A, param.ad);
	
	/*
	 * Evolve the quantum system until time t = T, where the steady state
         * is assumed to be reached.
	 * 
	 */
	printf("Start evolving system\n");
	if (dm_evolve_lme_t(qs, rho0, &param, rho_store_cb) != 0)
	//if (dm_evolve_he_t(qs, rho0, &param, rho_store_cb) != 0)
	//if (dm_evolve_unitary_const_full_steps(qs, rho0, &param, rho_store_cb) != 0)
	//if (dm_evolve_unitary_const_simple(qs, rho0, &param, rho_store_cb) != 0)
	//if (dm_evolve_unitary_const(qs, rho0, &param, rho_store_cb) != 0)
	//if (dm_evolve_unitary_t(qs, rho0, &param, rho_store_cb) != 0)
	{
		fprintf(stderr, "Evolution of the quantum system failed.\n");
		return -1;
	}

       // param.a  = operator_ho_lowering(qs, 1);
        //param.ad = operator_ho_raising(qs, 1);
        //param.A = gsl_matrix_complex_alloc(qsn, qsn);
        //gsl_matrix_complex_memcpy(param.A, param.a);
        //gsl_matrix_complex_add(param.A, param.ad);

	/* 
         * At this point the system (rho_f) is supposed to be in a steady-state,
	 * so that C(t,tau) = C(tau).
	 *
	 */
	rho_store_final_cb(param.rho_f, param.Tf, qs, &param);
	
	printf("Start calculating the correlation function\n");
        // here, assume that rho0 = rho(t), proceed to calculate Corr(t,tau)

        param.Ti =  0.00;
        param.Tf =  50.00;       // tau
        param.dT =  0.01;
        rho_t = gsl_matrix_complex_alloc(qsn, qsn);
        gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, GSL_COMPLEX_ONE, param.A,  param.rho_f,  GSL_COMPLEX_ZERO, rho_t);
        //gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, GSL_COMPLEX_ONE, param.A,  rho0,  GSL_COMPLEX_ZERO, rho_t);

        /*
         * Evolve the B rho(t)
         *
         */
        if (dm_evolve_lme_t(qs, rho_t, &param, rho_corr_cb) != 0)
	//if (dm_evolve_unitary_const(qs, rho_t, &param, rho_corr_cb) != 0)
        //if (dm_evolve_unitary_const(qs, rho, &param, rho_store_cb) != 0)
        //if (dm_evolve_unitary_t(qs, rho, &param, rho_store_cb) != 0)
        {
                fprintf(stderr, "Evolution of the B rho failed.\n");
                return -1;
        }
	
	return 0;
}



