/*
 * Simulate the time evolution of a system of N two-level systems (TLS).
 *
 * Robert Johansson <robert@riken.jp>
 *
 * $Id: run_qubits.c,v 1.1.1.1 2008-09-03 09:36:41 rob Exp $
 */
#include "quantum_system.h"
#include "density_matrix.h"
#include "operators.h"
#include "hamiltonian.h"
#include "master_equation.h"
#include "gsl_ext.h"
#include "param.h"
#include "misc.h"

#include <math.h>
#include <stdio.h> 
#include <gsl/gsl_matrix.h>


/**
 * Directory where all the data files are to be stored,
 * relative to the directory from which the program is
 * started.
 */
#define DATADIR "data/qubit.dynamics/"

/** 
 * Process the density matrix for time t (calc. expectations values, 
 * store to file, etc.). This is a callback function that the solver
 * call in each time-step.
 *
 * Here we can calculate expectation values etc. In this case we only
 * do simple 
 */
int
rho_store_cb(gsl_matrix_complex *rho_t, double t, quantum_system *qs, solver_param *sp)
{
	double c;
	int i;
	char filename[128], row[512];
	gsl_matrix_complex *dm_part;
	
        for (i = 0; i < qs->nsubsys; i++)
        {
		/*
		 * For each subsystem's density matrix, extract the diagonal elements
		 * that is the qubits occupation probabilities, and store to data
		 * files.
		 */
		snprintf(filename, sizeof(filename), "%s/qubit_population_%d.dat", DATADIR, i);
				
                dm_part = density_matrix_traceout(qs, rho_t, i);

		snprintf(row, sizeof(row), "%f\t%f\t%f", t, 
		         GSL_REAL(gsl_matrix_complex_get(dm_part, 0, 0)),	// rho^{qubit i}_{0,0}
		         GSL_REAL(gsl_matrix_complex_get(dm_part, 1, 1)));	// rho^{qubit i}_{1,1}
		
		misc_file_row_add(filename, row);
                gsl_matrix_complex_free(dm_part);
        }

	return 0;
}


/** 
 * Two-level system (qubit) parameters: 
 *
 * H = sum_i(-0.5 espilon[i] sigma_z(i) - 0.5 delta[i] sigma_x(i)) + sum_ij(- 0.5 lambda[i][j] sigma_x[i] sigma_x[j]) 
 * 
 */
double epsilon[] = {10.0, 10.0, 10.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0};
double delta[]   = { 0.0,  0.0,  0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0};
double *lambda[10];


/** 
 * Program starts here: Set up and start a calculation that simulates
 * the time-evolution of a coupled qubits.
 */
int
main(int argc, char **argv)
{
	quantum_system *qs;
	gsl_matrix_complex *rho1, *rho2;
	gsl_matrix_complex *rho0;
	density_matrix_list_t dm_list;
	int qsn, i;

	/* This data structure store all the parameters
	 * needed by the solver,
	 */
	solver_param param;

	/*
	 * Setup solver parameters: 
	 *
	 * Which parameters in solver_param that are used depends
	 * on which hamiltonian generating function that is used
	 * later on. Here we will use the hamiltonian for coupled
	 * two-level systems, so we configure the following parameters
	 * accordingly:
	 *
	 */

	/* Number of two-level systems (qubits) */
	param.N = 2;

	/* TLS hamiltonian parameters, i.e. 
	 * 
	 * 	- energy splitting: 	param.epsilon[i]
	 * 	- tunneling energy: 	param.delta[i]
	 * 	- inter-qubit coupling: lambda[i][j]
	 *
	 * where i and j is two qubit indices. The actual numerical
	 * values are defined in the arrays above.
	 */
	param.epsilon = epsilon;
	param.delta   = delta;
	for (i = 0; i < 10; i++)
	{
		param.epsilon[i] *= 2*M_PI;
		param.delta[i]   *= 2*M_PI;
		lambda[i] = (double *)calloc(10, sizeof(double));
	}
	/* qubits are coupled in a chain */
	lambda[0][1] = 0.100 * 2*M_PI;
	lambda[1][2] = 0.050 * 2*M_PI;
	param.lambda  = lambda;
	
	
	/* Configure the time range for the evolution:
	 *  - Ti = initial time
	 *  - Tf = final time
	 *  - dT = time-step between each call to the rho_store
	 *         callback function. The solver itself uses
	 *         and adaptive step length for accuracy control.
	 */
	param.Ti =   0.0;
	param.Tf =  20.0;
	param.dT =   0.01;

	/*
	 * If we want to apply a driving field to the qubits (i.e.
	 * for dynamic coupling, rabi oscillations, etc.), the following
	 * two variables define the amplitude and frequency of the 
	 * driving field. The operator that the driving field couple 
	 * to is defined by the hamiltonian_drive function, see below.
	 *
	 * Set the amplitude f_A to zero to disable driving.
	 */
	param.f_A = 1.5  * 2 * M_PI;
	param.f_w = 10.0 * 2 * M_PI;


	/* Function pointers to the functions that generate the Hamiltonian
	 * matrices. Ht_func is called by the solver and expected to return
	 * the instantanous value of the Hamiltonian (at time t). This function
	 * can be customized for each simulation, but often it is sufficient
	 * to use the standard from:
	 *
	 * H(t) = H_0 + H_1 * f_A * cos(f_w * t)
	 *
	 * which is implemented by the function hamiltonian_t from the 
	 * QDpack's library of standard Hamiltonians.
	 *
	 */
	param.H0_func = (hamiltonian_func_t)hamiltonian_nTLS;
	param.H1_func = (hamiltonian_func_t)hamiltonian_drive;
	param.Ht_func = (hamiltonian_td_func_t)hamiltonian_t;


	/* Create a new quantum system object: Here we configure
	 * the quantum systems that make up the total system. We
	 * call quantum_system_add for each subsystem, and the 
	 * second parameter defines how many quantum state that
	 * subsystem has.
	 *
	 */
	qs = quantum_system_new();
	for (i = 0; i < param.N; i++)
		quantum_system_add(qs, 2);
	qsn = quantum_system_nstates(qs);
	quantum_system_print(qs);


	/*
	 * Set up dissipation: Here we can define operators and rates 
	 * that appear in the master equation (relaxation rate, dephasing
	 * rates, etc.) for each subsystem.
	 *
	 */
	param.do_n = 0;

	// QUBIT 0:
	// no dissipation

	// QUBIT 1: 
	// Relaxation
	param.do_a[param.do_n] = operator_ho_lowering(qs, 1, 0); 	// qubit 1
	param.do_g1[param.do_n] = 0.075;   // relaxation rate
	param.do_g2[param.do_n] = 0.00;   // excitation rate (zero temperature)
	param.do_n++;
	// Pure dephasing
	// param.do_a[param.do_n] = operator_sigma_z(qs, 1, 0); 	// qubit 1
	// param.do_g1[param.do_n] = 0.25; // dephasing rate
	// param.do_n++;
	
	// QUBIT 2:
	// Relaxation
	//param.do_a[param.do_n] = operator_ho_lowering(qs, 2, 0); 	// qubit 2
	//param.do_g1[param.do_n] = 0.25;   // relaxation rate
	//param.do_g2[param.do_n] = 0.05;   // excitation rate 
	//param.do_n++;
	

	/*
	 * Setup initial state: Here we set the initial state for all qubits
	 * to its ground state.
	 */
	rho1 = density_matrix_pure_TLS(0.0); // = density matrix that correspond to |psi> = 1.0 |0> + 0.0 |1>
	rho2 = density_matrix_pure_TLS(0.0); // = density matrix that correspond to |psi> = 1.0 |0> + 0.0 |1> 
	
        density_matrix_list_init(&dm_list);
        density_matrix_list_append(&dm_list, rho1);
        for (i = 1; i < param.N; i++)
		density_matrix_list_append(&dm_list, rho2);

	rho0 = density_matrix_combine(qs, &dm_list);
	
	//printf("rho0 =\n");
	//misc_matrix_complex_print(rho0, qsn);


	/*
	 * Evolve the quantum system: We can use different solvers
	 * for different kind of evolution. Here we use the LME solver
	 * (Lindblad Master Equation).
	 *
	 * The solver will call the callback function "rho_store_cb" for
	 * each time-step. In the callback function we can calculate
	 * expectation values, etc., and store to files.
	 * 
	 */
	if (dm_evolve_lme_t(qs, rho0, &param, rho_store_cb) != 0)		// dissipative evolution: Lindblad Master Equation
	//if (dm_evolve_unitary_const(qs, rho0, &param, rho_store_cb) != 0)	// unitary evolution with constant Hamiltonian
	//if (dm_evolve_unitary_t(qs, rho0, &param, rho_store_cb) != 0)		// unitary evolution with time-dependent Hamiltonian
	{
		fprintf(stderr, "Evolution of the quantum system failed.\n");
		return -1;
	}

	return 0;
}



