/*
 * Simulate the time evolution of a N-TLS system.
 *
 *
 * Robert Johansson <robert@riken.jp>
 *
 * $Id: run_spinchain.c,v 1.1.1.1 2008-09-03 09:36:41 rob Exp $
 */
#include "quantum_system.h"
#include "density_matrix.h"
#include "entanglement.h"
#include "operators.h"
#include "hamiltonian.h"
#include "master_equation.h"
#include "gsl_ext.h"
#include "param.h"
#include "misc.h"

#include <math.h>
#include <stdio.h> 
#include <gsl/gsl_matrix.h>

/* ---
 * Process the density matrix for time t (calc. expectations values, 
 * store to file, etc.) 
 */
int
rho_store_cb(gsl_matrix_complex *rho_t, double t, quantum_system *qs, solver_param *sp)
{
	double ent_ln, ent_ln_sum = 0.0, ent_ne, ent_ne_sum = 0.0;
	int i, j, row1_offset, row2_offset;
	char filename1[128], filename2[128], row1[4096], row2[4096];
	gsl_matrix_complex *dm_part;
	quantum_system *qs_mask;

	qs_mask = quantum_system_copy(qs);
	for (i = 0; i < qs_mask->nsubsys; i++)
	{
		qs_mask->nstates[i] = 1;
	}

	snprintf(filename1, sizeof(filename1), "data.spinchain/ent_int_pairwise_%s.dat", sp->simsig);
	snprintf(filename2, sizeof(filename2), "data.spinchain/ent_ext_pairwise_%s.dat", sp->simsig);
	row1_offset = snprintf(row1, sizeof(row1), "%f", t);
	row2_offset = snprintf(row2, sizeof(row2), "%f", t);

	for (i = 0; i < qs->nsubsys-1; i++)
	{
		qs_mask->nstates[i] = 0;
		for (j = i+1; j < qs->nsubsys; j++)
		{
			qs_mask->nstates[j] = 0;
		
			dm_part = density_matrix_traceout_multi(qs, rho_t, qs_mask);
			ent_ln = entanglement_log_neg(dm_part);		
			ent_ln_sum += ent_ln;	
			row1_offset += snprintf(&row1[row1_offset], sizeof(row1)-row1_offset, "\t%f", ent_ln); 
			ent_ne = entanglement_neumann_entropy(dm_part);		
			ent_ne_sum += ent_ne;	
			row2_offset += snprintf(&row2[row2_offset], sizeof(row2)-row2_offset, "\t%f", ent_ne); 
		
			gsl_matrix_complex_free(dm_part);
			qs_mask->nstates[j] = 1;
		}
		qs_mask->nstates[i] = 1;
	}
	
	snprintf(&row1[row1_offset], sizeof(row1)-row1_offset, "\t%f", ent_ln_sum); 
	misc_file_row_add(filename1, row1);
	snprintf(&row2[row1_offset], sizeof(row2)-row2_offset, "\t%f", ent_ln_sum); 
	misc_file_row_add(filename2, row2);

	/* --- concurrence -- *
	snprintf(filename, sizeof(filename), "data.spinchain/concurrence.dat");
	c = density_matrix_concurrence(qs, rho_t);
	snprintf(row, sizeof(row), "%f\t%f", t, c); 
	misc_file_row_add(filename, row);
	*/

	return 0;
}


/* ---
 * Two-level system parameters: 
 *
 * H = sum_i(-0.5 espilon[i] sigma_z(i) - 0.5 delta[i] sigma_x(i)) + sum_ij(- 0.5 lambda[i][j] sigma_x[i] sigma_x[j]) 
 * 
 */
double epsilon[] = {0.3, 0.3, 0.3, 0.3, 0.3, 0.3, 0.3, 0.3, 0.3, 0.3,  0.3, 0.3, 0.3, 0.3, 0.3};
double delta[]   = {0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,  0.0, 0.0, 0.0, 0.0, 0.0};
double *lambda[15];


/* ---
 * Program starts here.
 * 
 */
int
main(int argc, char **argv)
{
	quantum_system *qs;
	gsl_matrix_complex *rho1, *rho2;
	gsl_matrix_complex *rho0;
	density_matrix_list_t dm_list;
	int qsn, i;

	solver_param param;

	/*
	 * Setup solver parameters.
	 *
	 */

	// TLS hamiltonian parameters
	param.epsilon = epsilon;
	param.delta   = delta;
	for (i = 0; i < 15; i++)
	{
		param.epsilon[i] *= 1;//2*M_PI;
		param.delta[i]   *= 1;//2*M_PI;
		lambda[i] = (double *)calloc(15, sizeof(double));
	}
	lambda[0][1] = 0.025;// * 2*M_PI;
	lambda[1][2] = 0.025;// * 2*M_PI;
	lambda[2][3] = 0.025;// * 2*M_PI;
	lambda[3][4] = 0.025;// * 2*M_PI;
	lambda[4][5] = 0.025;// * 2*M_PI;
	lambda[5][6] = 0.025;// * 2*M_PI;
	lambda[6][7] = 0.025;// * 2*M_PI;
	lambda[7][8] = 0.025;// * 2*M_PI;
	lambda[8][9] = 0.025;// * 2*M_PI;
	lambda[9][10] = 0.025;// * 2*M_PI;
	lambda[10][11] = 0.025;// * 2*M_PI;
	lambda[11][12] = 0.025;// * 2*M_PI;
	lambda[12][13] = 0.025;// * 2*M_PI;
	lambda[13][14] = 0.025;// * 2*M_PI;
	param.lambda  = lambda;
	
	
	// time range
	param.Ti =  0.0;
	param.Tf =  500.0;
	param.dT =  0.5;

	param.N = 12;

	// driving field (set to zero to disable)
	//param.f_A = 0.0 * 2 * M_PI;
	//param.f_w = 0.0 * 2 * M_PI;


	// function pointer to the functions generating the hamiltonian matrices
	param.H0_func = (hamiltonian_func_t)hamiltonian_nTLS;
	//param.H1_func = (hamiltonian_func_t)hamiltonian_drive;
	//param.Ht_func = (hamiltonian_td_func_t)hamiltonian_t;
	
	/* 
	 * create a new quantum system object
	 */
	qs = quantum_system_new();

	for (i = 0; i < param.N; i++)
		quantum_system_add(qs, 2);
	qsn = quantum_system_nstates(qs);
	quantum_system_print(qs);

	snprintf(param.simsig, sizeof(param.simsig), "_N_%d_w_%.2f_K_%.2f", param.N, epsilon[0], lambda[0][1]);

	/*
	 * Set up dissipation
	 *
	 *
	param.do_n = 0;
	param.do_a[param.do_n] = operator_ho_lowering(qs, 0); // qubit 0
	param.do_g1[param.do_n] = 0.50; // relaxation
	param.do_g2[param.do_n] = 0.05; // excitation
	param.do_n++;
	param.do_a[param.do_n] = operator_ho_lowering(qs, 1); // qubit 1
	param.do_g1[param.do_n] = 0.25; // relaxation
	param.do_g2[param.do_n] = 0.01; // excitation
	param.do_n++;
	param.do_a[param.do_n] = gsl_ext_matrix_convert_and_free(operator_sigma_z(qs, 0)); // qubit 0
	param.do_g1[param.do_n] = 0.25; // dephasing
	param.do_n++;
	*/

	/*
	 * debug
	 *
	 * /
	printf("sigma_x(0) = \n"); misc_matrix_print(operator_sigma_x(qs,0), 4);
	printf("sigma_x(0) = \n"); misc_matrix_print(operator_sigma_x(qs,1), 4);
	printf("sigma_y(0) = \n"); misc_matrix_complex_print(operator_sigma_y(qs,0), 4);
	printf("sigma_y(0) = \n"); misc_matrix_complex_print(operator_sigma_y(qs,1), 4);
	*/
	
	/*
	 * Setup initial state
	 */
	rho1 = density_matrix_pure_TLS(1.0); // 0.8|0>+0.2|1>
	rho2 = density_matrix_pure_TLS(1.0); // 0.3|0>+0.7|1> 
	
        density_matrix_list_init(&dm_list);
        density_matrix_list_append(&dm_list, rho1);
        for (i = 1; i < param.N; i++)
		density_matrix_list_append(&dm_list, rho2);

	rho0 = density_matrix_combine(qs, &dm_list);
	
	//printf("rho0 =\n");
	//misc_matrix_complex_print(rho0, qsn);


	/*
	 * Evolve the quantum system.
	 * 
	 */
	//if (dm_evolve_lme_t(qs, rho0, &param, rho_store_cb) != 0)
	if (dm_evolve_unitary_const(qs, rho0, &param, rho_store_cb) != 0)
	//if (dm_evolve_unitary_t(qs, rho0, &param, rho_store_cb) != 0)
	{
		fprintf(stderr, "Evolution of the quantum system failed.\n");
		return -1;
	}

	return 0;
}



