/*
 * Calculate the time evolution of the qubit + cavity + transmission
 * line system.
 *
 * Transmission spectrum calculation
 *
 * Robert Johansson <robert@riken.jp>
 *
 * $Id: run_tunable_ho.c,v 1.2 2008-11-04 02:49:32 rob Exp $
 */
#include "quantum_system.h"
#include "density_matrix.h"
#include "operators.h"
#include "hamiltonian.h"
#include "master_equation.h"
#include "gsl_ext.h"
#include "param.h"
#include "misc.h"

#include <math.h>
#include <stdio.h> 
#include <gsl/gsl_complex_math.h>
#include <gsl/gsl_complex.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_cblas.h>
#include <gsl/gsl_blas.h>


#define DATADIR "data/tun.ho"

#define WP_F 0.85
#define g  2.0*2*M_PI
#define Tsw	2.0

static double
ho_w_cb(double t, void *data)
{
	solver_param *sp = (solver_param *)data;
	double wp, wm, w;

	wm = sp->ho_w[0];
	wp = WP_F * wm;
	
	w = sqrt(wm*wm + (wp*wp-wm*wm) / (1 + exp(-2.0 * g * (t-Tsw))));
	
	return w;
}

/*
 * This function defines the time-dependence of the hamiltonian.
 *
 */
int
hamiltonian_sweep_t(gsl_matrix_complex *H_t, gsl_matrix_complex *H0, gsl_matrix_complex *H1, double t, solver_param *p)
{
	gsl_complex z;

        gsl_matrix_complex_memcpy(H_t, H0);
	GSL_SET_COMPLEX(&z, p->ho_w_func(t, p) / p->ho_w_func(0, p), 0.0); 
	gsl_matrix_complex_scale(H_t, z);

        return 0;
}


/* ---
 * Evaluate the correlation function <A(t+tau)A(t)> at tau
 *
 */
int
rho_corr_cb(gsl_matrix_complex *op_tau, double tau, quantum_system *qs, solver_param *sp)
{
        int qsn;
        gsl_complex corr1, corr2;
        gsl_matrix_complex *c;
        char filename[128], row[1024];

        qsn = quantum_system_nstates(qs);
        c = gsl_matrix_complex_alloc(qsn, qsn);

        gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, GSL_COMPLEX_ONE, sp->A,  op_tau,  GSL_COMPLEX_ZERO, c);

        corr1 = operator_trace(c);

        gsl_matrix_complex_free(c);

        corr2 = density_matrix_expectation_value(sp->A, op_tau);

        snprintf(filename, sizeof(filename), "%s/C%s.dat", DATADIR, sp->simsig);
        snprintf(row, sizeof(row), "%f\t%f\t%f\t%f\t%f", tau, GSL_REAL(corr1), GSL_IMAG(corr1), GSL_REAL(corr2), GSL_IMAG(corr2));
        misc_file_row_add(filename, row);

        return 0;
}

/* ---
 * Process the density matrix for time t (calc. expectations values, 
 * store to file, etc.) 
 */
int
rho_store_final_cb(gsl_matrix_complex *rho_t, double t, quantum_system *qs, solver_param *sp)
{
	gsl_complex N_expt; //, op_expt;
	char filename[128], row[1024];
	
	snprintf(filename, sizeof(filename), "%s/rho_f_expt_%d.dat", DATADIR, sp->N);
	
	/* -- expectation values -- */
	N_expt = density_matrix_expectation_value(rho_t, sp->N_op[1]);
	snprintf(row, sizeof(row), "%f\t%f\t%f\t%f", sp->f_w/(2*M_PI), sp->f_A/(2*M_PI), GSL_REAL(N_expt), GSL_IMAG(N_expt));
	misc_file_row_add(filename, row);		

	return 0;
}
	
/* ---
 * Process the density matrix for time t (calc. expectations values, 
 * store to file, etc.) 
 */
int
rho_store_cb(gsl_matrix_complex *rho_t, double t, quantum_system *qs, solver_param *sp)
{
	int i, j, ri;
	gsl_complex op_expt1, op_expt2, op_var; //N_expt
	char filename[128], row[1024];
	gsl_matrix_complex *dm_part; //, *op;

	//ri = 0;	
	//snprintf(filename, sizeof(filename), "%s/rho_diag%s.dat", DATADIR, sp->simsig);
	//ri = snprintf(row, sizeof(row), "%f", t);
	//for (j = 0; j < rho_t->size1; j++)
	//{
	// 	ri += snprintf(&row[ri], sizeof(row)-ri, "\t%f", (double)GSL_REAL(gsl_matrix_complex_get(rho_t, j, j)));
	//}
	//misc_file_row_add(filename, row);

	/* -- oscillator frequency -- */
        snprintf(filename, sizeof(filename), "%s/w_t%s.dat", DATADIR, sp->simsig);
        snprintf(row, sizeof(row), "%f\t%f", t, sp->ho_w_func(t, sp));
        misc_file_row_add(filename, row);
	
	for (i = 0; i < qs->nsubsys; i++)
        {
		ri = 0;	
		
		snprintf(filename, sizeof(filename), "%s/rho_ho_%d_diag%s.dat", DATADIR, i, sp->simsig);
		
                dm_part = density_matrix_traceout(qs, rho_t, i);

		ri = snprintf(row, sizeof(row), "%f", t);

		for (j = 0; j < qs->nstates[i]; j++)
		{
		 	ri += snprintf(&row[ri], sizeof(row)-ri, "\t%f", (double)GSL_REAL(gsl_matrix_complex_get(dm_part, j, j)));
		}
		misc_file_row_add(filename, row);
		
		/* -- expectation values -- * / 
		N_expt = density_matrix_expectation_value(rho_t, sp->N_op[i]);
		snprintf(filename, sizeof(filename), "%s/N_expt_%d%s.dat", DATADIR, i, sp->simsig);
		snprintf(row, sizeof(row), "%f\t%f\t%f", t, GSL_REAL(N_expt), GSL_IMAG(N_expt));
		misc_file_row_add(filename, row);		
		*/

		
		/* -- trace check values -- * /
		op = gsl_ext_matrix_convert_and_free(operator_unit(qs, i));
		N_expt = density_matrix_expectation_value(dm_part, op);
		snprintf(filename, sizeof(filename), "%s/Trace_%d%s.dat", DATADIR, i, sp->simsig);
		snprintf(row, sizeof(row), "%f\t%f\t%f", t, GSL_REAL(N_expt), GSL_IMAG(N_expt));
		misc_file_row_add(filename, row);		
		gsl_matrix_complex_free(op);
		*/	
		
/*
		op_expt = density_matrix_expectation_value(rho_t, sp->a);
		snprintf(filename, sizeof(filename), "%s/a_expt_%d%s.dat", DATADIR, i, sp->simsig);
		snprintf(row, sizeof(row), "%f\t%f\t%f", t, GSL_REAL(op_expt), GSL_IMAG(op_expt));
		misc_file_row_add(filename, row);		

		op_expt = density_matrix_expectation_value(rho_t, sp->ad);
		snprintf(filename, sizeof(filename), "%s/ad_expt_%d%s.dat", DATADIR, i, sp->simsig);
		snprintf(row, sizeof(row), "%f\t%f\t%f", t, GSL_REAL(op_expt), GSL_IMAG(op_expt));
		misc_file_row_add(filename, row);		

		op_expt = density_matrix_expectation_value(rho_t, sp->b);
		snprintf(filename, sizeof(filename), "%s/b_expt_%d%s.dat", DATADIR, i, sp->simsig);
		snprintf(row, sizeof(row), "%f\t%f\t%f", t, GSL_REAL(op_expt), GSL_IMAG(op_expt));
		misc_file_row_add(filename, row);		

		op_expt = density_matrix_expectation_value(rho_t, sp->bd);
		snprintf(filename, sizeof(filename), "%s/bd_expt_%d%s.dat", DATADIR, i, sp->simsig);
		snprintf(row, sizeof(row), "%f\t%f\t%f", t, GSL_REAL(op_expt), GSL_IMAG(op_expt));
		misc_file_row_add(filename, row);		

		op_expt = density_matrix_expectation_value(rho_t, sp->b2);
		snprintf(filename, sizeof(filename), "%s/b2_expt_%d%s.dat", DATADIR, i, sp->simsig);
		snprintf(row, sizeof(row), "%f\t%f\t%f", t, GSL_REAL(op_expt), GSL_IMAG(op_expt));
		misc_file_row_add(filename, row);		

		op_expt = density_matrix_expectation_value(rho_t, sp->bd2);
		snprintf(filename, sizeof(filename), "%s/bd2_expt_%d%s.dat", DATADIR, i, sp->simsig);
		snprintf(row, sizeof(row), "%f\t%f\t%f", t, GSL_REAL(op_expt), GSL_IMAG(op_expt));
		misc_file_row_add(filename, row);		
*/
		
		//gsl_matrix_complex_free(dm_part);
        }

	/* number operator expectation value, variance and the Q-parameter */
	
	/* -- n -- */
	op_expt1 = density_matrix_expectation_value(rho_t, sp->n_op);
	snprintf(filename, sizeof(filename), "%s/n_expt%s.dat", DATADIR, sp->simsig);
	snprintf(row, sizeof(row), "%f\t%f\t%f", t, GSL_REAL(op_expt1), GSL_IMAG(op_expt1));
	misc_file_row_add(filename, row);		
	
	op_expt2 = density_matrix_expectation_value(rho_t, sp->n_op_2);
	//snprintf(filename, sizeof(filename), "%s/n_2_expt%s.dat", DATADIR, sp->simsig);
	//snprintf(row, sizeof(row), "%f\t%f\t%f", t, GSL_REAL(op_expt2), GSL_IMAG(op_expt2));
	//misc_file_row_add(filename, row);		
	
	op_var = gsl_complex_sub(op_expt2, gsl_complex_mul(op_expt1, op_expt1));
	snprintf(filename, sizeof(filename), "%s/n_var%s.dat", DATADIR, sp->simsig);
	snprintf(row, sizeof(row), "%f\t%f\t%f", t, GSL_REAL(op_var), GSL_IMAG(op_var));
	misc_file_row_add(filename, row);	

	op_var = gsl_complex_sub_real(gsl_complex_div(op_var, op_expt1), 1);
	snprintf(filename, sizeof(filename), "%s/Q_param%s.dat", DATADIR, sp->simsig);
	snprintf(row, sizeof(row), "%f\t%f\t%f", t, GSL_REAL(op_var), GSL_IMAG(op_var));
	misc_file_row_add(filename, row);	
	
	
	/* Quadrature expectation values and variances */
	
	/* -- x1 -- */
	op_expt1 = density_matrix_expectation_value(rho_t, sp->x1);
	snprintf(filename, sizeof(filename), "%s/x1_expt%s.dat", DATADIR, sp->simsig);
	snprintf(row, sizeof(row), "%f\t%f\t%f", t, GSL_REAL(op_expt1), GSL_IMAG(op_expt1));
	misc_file_row_add(filename, row);		
	
	op_expt2 = density_matrix_expectation_value(rho_t, sp->x1_2);
	//snprintf(filename, sizeof(filename), "%s/x1_2_expt%s.dat", DATADIR, sp->simsig);
	//snprintf(row, sizeof(row), "%f\t%f\t%f", t, GSL_REAL(op_expt2), GSL_IMAG(op_expt2));
	//misc_file_row_add(filename, row);		
	
	op_var = gsl_complex_sub(op_expt2, gsl_complex_mul(op_expt1, op_expt1));
	snprintf(filename, sizeof(filename), "%s/x1_var%s.dat", DATADIR, sp->simsig);
	snprintf(row, sizeof(row), "%f\t%f\t%f", t, GSL_REAL(op_var), GSL_IMAG(op_var));
	misc_file_row_add(filename, row);	


	/* -- x2 -- */
	op_expt1 = density_matrix_expectation_value(rho_t, sp->x2);
	snprintf(filename, sizeof(filename), "%s/x2_expt%s.dat", DATADIR, sp->simsig);
	snprintf(row, sizeof(row), "%f\t%f\t%f", t, GSL_REAL(op_expt1), GSL_IMAG(op_expt1));
	misc_file_row_add(filename, row);		
	
	op_expt2 = density_matrix_expectation_value(rho_t, sp->x2_2);
	//snprintf(filename, sizeof(filename), "%s/x2_2_expt%s.dat", DATADIR, sp->simsig);
	//snprintf(row, sizeof(row), "%f\t%f\t%f", t, GSL_REAL(op_expt2), GSL_IMAG(op_expt2));
	//misc_file_row_add(filename, row);		
	
	op_var = gsl_complex_sub(op_expt2, gsl_complex_mul(op_expt1, op_expt1));
	snprintf(filename, sizeof(filename), "%s/x2_var%s.dat", DATADIR, sp->simsig);
	snprintf(row, sizeof(row), "%f\t%f\t%f", t, GSL_REAL(op_var), GSL_IMAG(op_var));
	misc_file_row_add(filename, row);	
	
		
	/* -- expectation values -- * /
	N_expt = density_matrix_expectation_value(rho_t, sp->A);
	snprintf(filename, sizeof(filename), "%s/E_expt%s.dat", DATADIR, sp->simsig);
	snprintf(row, sizeof(row), "%f\t%f\t%f", t, GSL_REAL(N_expt), GSL_IMAG(N_expt));
	misc_file_row_add(filename, row);		
	*/
		
	/* -- trace check values -- * /
	op = gsl_ext_matrix_convert_and_free(operator_unit(qs, -1));
	N_expt = density_matrix_expectation_value(rho_t, op);
	snprintf(filename, sizeof(filename), "%s/Trace_full%s.dat", DATADIR, sp->simsig);
	snprintf(row, sizeof(row), "%f\t%f\t%f", t, GSL_REAL(N_expt), GSL_IMAG(N_expt));
	misc_file_row_add(filename, row);		
	gsl_matrix_complex_free(op);
	*/
	
//	if (t >= sp->Tf-sp->dT)
//	{
//		printf("Storing final rho: rho_f\n");
//	 	gsl_matrix_complex_memcpy(sp->rho_f, rho_t);
//	}
	
	return 0;
}


/* ---
 * Simulation parameters.
 * 
 */
//double epsilon[] = { 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0};
//double delta[]   = { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0};
double ho_w[]    = { 1.0,  0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0};
double *lambda[10];


/* ---
 * Execution starts here.
 * 
 */
int
main(int argc, char **argv)
{
	quantum_system *qs;
	gsl_matrix_complex *rho_c;
	//gsl_matrix_complex *rho_q;
	gsl_matrix_complex *rho0, *rho_t;
	density_matrix_list_t dm_list;
	int qsn, i;
	double Navg, fw, fA, Gc, dummy, theta, phi_param;
	gsl_complex alpha;

	solver_param param;

	if (argc != 6)
	{
		printf("usage: %s check code\n", argv[0]);
		exit(0);
	}
	
	dummy   = strtod(argv[1], NULL);
	fw      = strtod(argv[2], NULL);
	fA      = strtod(argv[3], NULL);
	dummy	= strtod(argv[4], NULL);
	dummy	= strtod(argv[5], NULL);
	phi_param = 0.0;

	//printf("Input arguments: phi = %f\n", phi);
	
	/*
	 * System parameters.
	 *
	 */
	
	//the qubit
	//epsilon[0] = 2.899;
	//delta[0] = 0;

	//ho_w[0] = 1/(2*M_PI);
	
	//param.epsilon = epsilon;
	//param.delta   = delta;
	param.ho_w    = ho_w;
	for (i = 0; i < 10; i++)
	{
		//param.epsilon[i] *= 2*M_PI;
		//param.delta[i]   *= 2*M_PI;
		param.ho_w[i]    *= 2*M_PI;
		//lambda[i] = (double *)calloc(10, sizeof(double));
	}


	/*
	 * Coupling strengths
	 *
	 */
	//lambda[0][1] = 0.0 * 2*M_PI;
	//param.lambda  = lambda;
	//param.g_qc = gqc * 2*M_PI; 
	//param.g_cr = 0.0 * 2*M_PI;

	/*
	 * Driving parameters
	 *
	 */
	//fA = 0.25;
	//fw = 1.0;
	//param.f_A = fA * 2 * M_PI;
	//param.f_w = fw * ho_w[0];
	
	/* 
	 * Decay rates 
	 *
	 */
	Gc = 0.00;	// cavity decay rate
	//Gq = 1.0;	// qubit decay rate

	/*
	 * Time evolution parameters
	 */
	param.Ti =  0.0;
	param.Tf =  6.0;
	param.dT =  0.01;

	param.N = 25;
	param.n_offset = 0;
	//param.N = (int)(Gq/(2*Gc)) + 5;

	param.H0_func = (hamiltonian_func_t)hamiltonian_ho;
	param.H1_func = NULL; //(hamiltonian_func_t)hamiltonian_ho_drive;
	param.Ht_func = (hamiltonian_td_func_t)hamiltonian_sweep_t;

	/* 
	 * create a new quantum system object
	 */
	qs = quantum_system_new();
	quantum_system_add(qs, param.N);
	qsn = quantum_system_nstates(qs);
	quantum_system_print(qs);
	param.qs = qs;

	//param.H0 = hamiltonian_ho(param.qs, &param);
	//printf("H_real =\n");
	//misc_matrix_complex_print_real(param.H0, param.N);
	//printf("H_imag =\n");
	//misc_matrix_complex_print_imag(param.H0, param.N);
	//exit(0);

	/*
	 * Configure continous basis transform.
	 *
	 */
	param.cont_basis_transform = 1;
	param.ho_w_func = (ho_w_func_t)ho_w_cb;

	
	/*
	 * Set up dissipation
	 *
	 */
	param.wT = 0.0 * 2 * M_PI;
	param.do_n = 0;
	Navg = 0;

	/* --- qubit --- * /
	//param.do_a[param.do_n] = operator_ho_lowering(qs, 0, 0); 
	param.do_a[param.do_n] = operator_ho_raising(qs, 0, 0); 
	Navg = 1/(exp(sqrt(param.epsilon[0]*param.epsilon[0] + param.delta[0]*param.delta[0])/param.wT) - 1);
	printf("qubit Navg = %f\n", Navg);
	param.do_g1[param.do_n] = Gq * (1 + Navg); // relaxation
	param.do_g2[param.do_n] = Gq * Navg;       // excitation
	param.do_n++;
	*/
	
	/* --- cavity --- * /
 	param.do_a[param.do_n] = operator_ho_lowering(qs, 0, 0); 
	Navg = 1/(exp(param.ho_w[0]/param.wT) - 1);
	param.do_g1[param.do_n] = Gc * (1 + Navg); // relaxation
	param.do_g2[param.do_n] = Gc * Navg;       // excitation
	param.do_n++;
	*/
	
	/* --- transmission line --- * /
 	param.do_a[param.do_n] = operator_ho_lowering(qs, 2, 0); 
	Navg = 1/(exp(param.ho_w[1]/param.wT) - 1);
	param.do_g1[param.do_n] = 0.05 * (1 + Navg); // relaxation
	param.do_g2[param.do_n] = 0.05 * Navg;       // excitation
	param.do_n++;
	*/

	/* 
	 * Pre-calc of operators to calculate expectation values for.
	 */
	param.n_op = operator_ho_N(qs, 0, 0);
	param.n_op_2  = gsl_matrix_complex_alloc(qsn, qsn);
	gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, GSL_COMPLEX_ONE, param.n_op, param.n_op, GSL_COMPLEX_ZERO, param.n_op_2);
	
	param.a  = operator_ho_lowering(qs, 0, 0);
        param.ad = operator_ho_raising(qs, 0, 0);

	// Quandrature operators
	param.x1 = gsl_matrix_complex_alloc(qsn, qsn);
        gsl_matrix_complex_memcpy(param.x1, param.a);
        gsl_matrix_complex_add(param.x1, param.ad);
	GSL_SET_COMPLEX(&alpha, 0.5, 0.0);
	gsl_matrix_complex_scale(param.x1, alpha);
	param.x1_2  = gsl_matrix_complex_alloc(qsn, qsn);
	gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, GSL_COMPLEX_ONE, param.x1, param.x1, GSL_COMPLEX_ZERO, param.x1_2);
	
	param.x2 = gsl_matrix_complex_alloc(qsn, qsn);
        gsl_matrix_complex_memcpy(param.x2, param.a);
        gsl_matrix_complex_sub(param.x2, param.ad);
	GSL_SET_COMPLEX(&alpha, 0.0, -0.5);
	gsl_matrix_complex_scale(param.x2, alpha);
	param.x2_2  = gsl_matrix_complex_alloc(qsn, qsn);
	gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, GSL_COMPLEX_ONE, param.x2, param.x2, GSL_COMPLEX_ZERO, param.x2_2);
	
        //param.A  = gsl_matrix_complex_alloc(qsn, qsn);
        //gsl_matrix_complex_memcpy(param.A, param.a);
        //gsl_matrix_complex_add(param.A, param.ad);

/*
	param.b  = operator_ho_lowering(qs, 1);
	param.bd = operator_ho_raising(qs, 1);
	param.b2  = gsl_matrix_complex_alloc(qsn, qsn);
	param.bd2  = gsl_matrix_complex_alloc(qsn, qsn);
	{
		gsl_complex alpha, beta;
		GSL_SET_COMPLEX(&alpha, 1.0, 0.0);
		GSL_SET_COMPLEX(&beta, 0.0, 0.0);
		gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, param.b,  param.b,  beta, param.b2);
		gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, param.bd, param.bd, beta, param.bd2);
	}
*/
		
	snprintf(param.simsig, sizeof(param.simsig), "_Nc_%d_wc_%.3f_Gc_%.3f",
		qs->nstates[0], param.ho_w[0]/(2*M_PI), Gc);

	/*
	 * Setup initial state
	 */
	//rho_q = density_matrix_pure_TLS(1.0); 		// |> state
	//rho_c = density_matrix_fock_state(5, qs->nstates[0]); 	// |0> state
	theta = param.ho_w[0] * Tsw + (phi_param + 0.25) * 2 * M_PI;
	rho_c = density_matrix_coherent_state(sqrt(3), theta, qs->nstates[0], 0); 	// |0> state
	
        density_matrix_list_init(&dm_list);
        //density_matrix_list_append(&dm_list, rho_q);
        density_matrix_list_append(&dm_list, rho_c);
	rho0 = density_matrix_combine(qs, &dm_list);
       
	
	/* storage space for final rho */
	//param.rho_f = gsl_matrix_complex_alloc(qsn, qsn);

	/*
	 * Evolve the quantum system until time t = T, where the steady state
         * is assumed to be reached.
	 * 
	 */
	printf("Start evolving system\n");
	if (dm_evolve_lme_t(qs, rho0, &param, rho_store_cb) != 0)
	//if (dm_evolve_he_t(qs, rho0, &param, rho_store_cb) != 0)
	//if (dm_evolve_unitary_const_full_steps(qs, rho0, &param, rho_store_cb) != 0)
	//if (dm_evolve_unitary_const_simple(qs, rho0, &param, rho_store_cb) != 0)
	//if (dm_evolve_unitary_const(qs, rho0, &param, rho_store_cb) != 0)
	//if (dm_evolve_unitary_t(qs, rho0, &param, rho_store_cb) != 0)
	{
		fprintf(stderr, "Evolution of the quantum system failed.\n");
		return -1;
	}

	exit(0);
	
	/* 
         * At this point the system (rho_f) is supposed to be in a steady-state,
	 * such that C(t,tau) = C(tau).
	 *
	 */
	rho_store_final_cb(param.rho_f, param.Tf, qs, &param);
	
	printf("Start calculating the correlation function\n");
        // here, assume that rho0 = rho(t), proceed to calculate Corr(t,tau)

        //param.Ti =  0.00;
        //param.Tf =  1000.00;       // tau
        //param.dT =  0.5;

        rho_t = gsl_matrix_complex_alloc(qsn, qsn);
        gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, GSL_COMPLEX_ONE, param.A,  param.rho_f,  GSL_COMPLEX_ZERO, rho_t);
        //gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, GSL_COMPLEX_ONE, param.A,  rho0,  GSL_COMPLEX_ZERO, rho_t);

        /*
         * Evolve the B rho(t)
         *
         */
        if (dm_evolve_lme_t(qs, rho_t, &param, rho_corr_cb) != 0)
	//if (dm_evolve_unitary_const(qs, rho_t, &param, rho_corr_cb) != 0)
        //if (dm_evolve_unitary_const(qs, rho, &param, rho_store_cb) != 0)
        //if (dm_evolve_unitary_t(qs, rho, &param, rho_store_cb) != 0)
        {
                fprintf(stderr, "Evolution of the B rho failed.\n");
                return -1;
        }
	
	return 0;
}



